"""Test module for pyscal.*.add_fromtable()"""

import hypothesis.strategies as st
import numpy as np
import pandas as pd
import pytest
from hypothesis import given

from pyscal import GasOil, WaterOil, WaterOilGas
from pyscal.utils.testing import check_table, float_df_checker


def test_wo_fromtable_simple():
    """Test loading a simple curve from a table"""
    df1 = pd.DataFrame(
        columns=["SW", "KRW", "KROW", "PC"], data=[[0, 0, 1, 2], [1, 1, 0, 0]]
    )
    wateroil = WaterOil(h=0.1)
    # With wrong names:
    with pytest.raises(ValueError):
        wateroil.add_fromtable(df1, swcolname="sw")

    # Set names:
    wateroil.add_fromtable(df1, swcolname="SW", pccolname="PC")
    assert "KRW" in wateroil.table.columns
    assert "KROW" in wateroil.table.columns
    assert "PC" in wateroil.table.columns
    assert sum(wateroil.table["KRW"]) > 0
    assert sum(wateroil.table["KROW"]) > 0
    assert np.isclose(sum(wateroil.table["PC"]), 11)  # Linearly increasing PC
    check_table(wateroil.table)


def test_go_fromtable_simple():
    """Test reading of a simple gasoil table"""
    df1 = pd.DataFrame(
        columns=["SG", "KRG", "KROG", "PC"], data=[[0, 0, 1, 0], [1, 1, 0, 2]]
    )
    gasoil = GasOil(h=0.1)
    gasoil.add_fromtable(
        df1, sgcolname="SG", krgcolname="KRG", krogcolname="KROG", pccolname="PC"
    )
    assert sum(gasoil.table["KRG"]) > 0
    assert sum(gasoil.table["KROG"]) > 0
    assert np.isclose(sum(gasoil.table["PC"]), 11)  # Linearly increasing PCOG
    check_table(gasoil.table)

    # Check that we can read the table also when the dataframe dtypes are strings
    gasoil_str = GasOil(h=0.1)
    gasoil_str.add_fromtable(
        df1.astype(str),
        sgcolname="SG",
        krgcolname="KRG",
        krogcolname="KROG",
        pccolname="PC",
    )
    assert sum(gasoil_str.table["KRG"]) > 0
    assert sum(gasoil_str.table["KROG"]) > 0
    check_table(gasoil_str.table)


def test_wo_fromtable_multiindex():
    """Test that we accept multiindex dataframes,
    (but a warning will be issued)"""
    # Test an example dataframe that easily gets sent in from ecl2df.satfunc:
    dframe = pd.DataFrame(
        columns=["KEYWORD", "SATNUM", "SW", "KRW", "KROW", "PC"],
        data=[
            ["SWOF", 1, 0, 0, 1, 2],
            ["SWOF", 1, 0.5, 0.5, 0.5, 1],
            ["SWOF", 1, 1, 1, 0, 0],
        ],
    ).set_index(["KEYWORD", "SATNUM"])

    # Check that we have a MultiIndex:
    assert len(dframe.index.names) == 2

    wateroil = WaterOil(h=0.1)
    wateroil.add_fromtable(
        dframe, swcolname="SW", krwcolname="KRW", krowcolname="KROW", pccolname="PC"
    )
    assert "KRW" in wateroil.table.columns
    assert "KROW" in wateroil.table.columns
    assert "PC" in wateroil.table.columns
    check_table(wateroil.table)


def test_go_fromtable_multiindex():
    """Test that we accept multiindex dataframes,
    (but a warning will be issued)"""
    # Test an example dataframe that easily gets sent in from ecl2df.satfunc:
    dframe = pd.DataFrame(
        columns=["KEYWORD", "SATNUM", "SG", "KRG", "KROG", "PC"],
        data=[
            ["SGOF", 1, 0, 0, 1, 0],
            ["SGOF", 1, 0.5, 0.5, 0.5, 0],
            ["SGOF", 1, 1, 1, 0, 0],
        ],
    ).set_index(["KEYWORD", "SATNUM"])

    # Check that we have a MultiIndex:
    assert len(dframe.index.names) == 2

    gasoil = GasOil(h=0.1)
    gasoil.add_fromtable(
        dframe, sgcolname="SG", krgcolname="KRG", krogcolname="KROG", pccolname="PC"
    )
    assert "KRG" in gasoil.table.columns
    assert "KROG" in gasoil.table.columns
    assert "PC" in gasoil.table.columns
    check_table(gasoil.table)


def test_go_fromtable_problems():
    """Test loading from a table where there should be problems"""
    df1 = pd.DataFrame(
        columns=["SG", "KRG", "KROG", "PCOG"], data=[[0.1, 0, 1, 0], [0.9, 1, 0, 2]]
    )
    # Now sgcr and swl is wrong:
    gasoil = GasOil(h=0.1)

    with pytest.raises(ValueError, match="SBOGUS not found in dataframe"):
        gasoil.add_fromtable(df1, sgcolname="SBOGUS")

    df2 = pd.DataFrame(
        columns=["SG", "KRG", "KROG", "PCOG"], data=[[0.0, 0, 1, 0], [0.9, 0.8, 0, 2]]
    )
    with pytest.raises(ValueError):
        # should say too large swl for pcog interpolation
        gasoil = GasOil(h=0.1)
        gasoil.add_fromtable(df2, pccolname="PCOG")
    gasoil = GasOil(h=0.1, swl=0.1)
    gasoil.add_fromtable(df2, pccolname="PCOG")
    assert np.isclose(gasoil.table["PC"].max(), 2.0)
    assert np.isclose(gasoil.table["PC"].min(), 0.0)
    assert np.isclose(gasoil.table["SG"].max(), 0.9)

    gasoil = GasOil(h=0.1)
    with pytest.raises(ValueError):
        # sg must start at zero
        gasoil.add_fromtable(df1, krgcolname="KRG", krogcolname="KROG")
    # This works fine, but we get warnings on swl not seemingly correct
    gasoil.add_fromtable(df2, krgcolname="KRG", krogcolname="KROG", pccolname=None)
    # krg will be extrapolated to sg=1
    float_df_checker(gasoil.table, "SG", 1.0, "KRG", 0.8)
    float_df_checker(gasoil.table, "SG", 1.0, "KROG", 0.0)
    gasoil = GasOil(h=0.1, swl=0.1)
    gasoil.add_fromtable(df2)


@pytest.mark.parametrize(
    "dframe, exception, message",
    [
        (
            pd.DataFrame(
                columns=["SG", "KRG", "KROG", "PCOG"],
                data=[[0, -0.01, 1, 0], [1, 1, 0, 0]],
            ),
            ValueError,
            "krg is below 0 in incoming table",
        ),
        (
            pd.DataFrame(
                columns=["SG", "KRG", "KROG", "PCOG"],
                data=[[0, 0, 1, 0], [1, 1.0000000001, 0, 0]],
            ),
            ValueError,
            "krg is above 1 in incoming table",
        ),
        (
            pd.DataFrame(
                columns=["SG", "KRG", "KROG", "PCOG"],
                data=[[0, 0, 0.5, 0], [1, 1, 0.9, 0]],
            ),
            ValueError,
            "Incoming krog not decreasing",
        ),
        (
            pd.DataFrame(
                columns=["SG", "KRG", "KROG", "PCOG"],
                data=[[0, 0, 1, 0], [1, 1, -0.1, 0]],
            ),
            ValueError,
            "krog is below 0 in incoming table",
        ),
        (
            pd.DataFrame(
                columns=["SG", "KRG", "KROG", "PCOG"],
                data=[[0.1, 0, 1, 0], [1, 1, 0.1, 0]],
            ),
            ValueError,
            "sg must start at zero",
        ),
        (
            pd.DataFrame(
                columns=["SG", "KRG", "KROG", "PCOG"],
                data=[[0.0, 0, 1, 1], [1, 1, 0.1, 0]],
            ),
            ValueError,
            "Incoming pc not increasing",
        ),
        (
            pd.DataFrame(
                columns=["SG", "KRG", "KROG", "PCOG"],
                data=[[0.0, 0, 1, 0], [1, 1, 0.1, np.inf]],
            ),
            ValueError,
            # This one is from scipy/interpolate/_cubic.py
            "must contain at least 2 elements",
        ),
        (
            pd.DataFrame(
                columns=["SG", "KRG", "KROG", "PCOG"],
                data=[[0.0, 0, 1, 0], [1, 1, 0.1, np.nan]],
            ),
            ValueError,
            # This one is from scipy/interpolate/_cubic.py
            "must contain at least 2 elements",
        ),
        (
            pd.DataFrame(
                columns=["SG", "KRG", "KROG", "PCOG"],
                data=[[0.0, 0, 1, 0], [0.5, 0.5, 0.5, 0.5], [1, 1, 0.1, np.inf]],
            ),
            ValueError,
            "inf/nan in interpolated data",
        ),
    ],
)
def test_go_from_table_exceptions(dframe, exception, message):
    """Test exceptions in loading tabular data for gasoil"""
    gasoil = GasOil(h=0.1)
    with pytest.raises(exception, match=message):
        gasoil.add_fromtable(dframe)


def test_gascondensate():
    """Test how sgro works when data is tabulated. Sgro is to be used
    for gas condensate modelling. sgro is tricky because only 0 and sgcr
    are valid values for sgro when constructing the objects."""

    # When sgro is nonzero, we are doing gas condensate modelling:
    gascond_orig = GasOil(sgro=0.2, sgcr=0.2, h=0.1)
    gascond_orig.add_corey_gas()
    gascond_orig.add_corey_oil(kroend=0.8)

    # Make a new object without assuming anything about sgro and sgcr:
    gascond_tabulated = GasOil(h=0.1)
    gascond_tabulated.add_fromtable(gascond_orig.table)
    check_table(gascond_tabulated.table)

    # sgro is estimated correctly in this case:
    assert np.isclose(gascond_tabulated.sgro, gascond_orig.sgro)

    # The object constructed from the table has an extra row at SG=0.1, because
    # we didn't tell it that sgcr was 0.2, otherwise they are numerically
    # equivalent:
    assert len(gascond_tabulated.table) == len(gascond_orig.table) + 1
    pd.testing.assert_frame_equal(
        gascond_orig.table.iloc[1:][["SG", "KRG", "KROG"]].reset_index(drop=True),
        gascond_tabulated.table.iloc[2:][["SG", "KRG", "KROG"]].reset_index(drop=True),
    )

    # Make a tricky gascondensate object which has a linear oil curve. The sgro
    # estimate will become 1.0, we should still be able to use this as a table.
    gascond_linear = GasOil(sgro=0.2, sgcr=0.2, h=0.1)
    gascond_linear.add_corey_gas()
    gascond_linear.add_corey_oil(kroend=0.8, kromax=1, nog=1)
    gascond_linear_tabulated = GasOil(h=0.1)
    gascond_linear_tabulated.add_fromtable(gascond_linear.table)

    check_table(gascond_linear_tabulated.table)

    # sgro could not be guessed here, and is reset to zero:
    assert np.isclose(gascond_linear_tabulated.sgro, 0.0)


def test_wo_singlecolumns():
    """Test that we can load single columns from individual dataframes"""
    krw = pd.DataFrame(columns=["SW", "KRW"], data=[[0.15, 0], [0.89, 1], [1, 1]])
    krow = pd.DataFrame(columns=["SW", "KROW"], data=[[0.15, 1], [0.89, 0], [1, 0]])
    pc1 = pd.DataFrame(columns=["SW", "PCOW"], data=[[0.15, 3], [0.89, 0.1], [1, 0]])
    wateroil = WaterOil(h=0.1, swl=0.15, sorw=1 - 0.89)
    wateroil.add_fromtable(krw)
    assert "KRW" in wateroil.table
    assert "KROW" not in wateroil.table
    wateroil.add_fromtable(krow)
    assert "KROW" in wateroil.table
    wateroil.add_fromtable(pc1)
    assert "PC" in wateroil.table

    # We want to allow a pc dataframe where sw starts from zero:
    # then should not preserve pc(sw=0)
    pc2 = pd.DataFrame(columns=["SW", "PCOW"], data=[[0, 3], [0.5, 0.1], [1, 0]])
    wateroil.add_fromtable(pc2)
    assert "PC" in wateroil.table
    assert wateroil.table["SW"].min() == 0.15
    assert wateroil.table["PC"].max() < 3

    # But we should refuse a pc dataframe not covering our sw range:
    wateroil = WaterOil(h=0.1, swl=0)
    with pytest.raises(ValueError):
        pc3 = pd.DataFrame(columns=["SW", "PCOW"], data=[[0.1, 3], [0.5, 0.1], [1, 0]])
        wateroil.add_fromtable(pc3)
    with pytest.raises(ValueError):
        pc3 = pd.DataFrame(columns=["SW", "PCOW"], data=[[0, 3], [0.5, 0.1]])
        wateroil.add_fromtable(pc3)

    # Disallow non-monotonous capillary pressure:
    pc4 = pd.DataFrame(columns=["SW", "PCOW"], data=[[0.15, 3], [0.89, 0.1], [1, 0.1]])
    with pytest.raises(ValueError):
        wateroil.add_fromtable(pc4)

    # But if capillary pressure is all zero, accept it:
    pc5 = pd.DataFrame(columns=["SW", "PCOW"], data=[[0, 0], [1, 0]])
    wateroil = WaterOil(h=0.1)
    wateroil.add_fromtable(pc5)
    assert wateroil.table["PC"].sum() == 0


@pytest.mark.parametrize(
    "h, sw_mid",
    [
        (0.1, 0.1),
        (0.2, 0.1),
        (0.1, 0.2),
        (0.1, 0.1000001),
        (0.1, 0.9999999),
        (1, 0.1),
    ],
)
def test_linear_input(h, sw_mid):
    """Linear input creates difficulties with sorw, which is used in
    add_fromtable(). The intention of the test is to avoid crashes when
    add_fromtable().

    estimate_sorw() is unreliable on linear input, and returns 0 or 1 on the
    given test dataset. Correctness of sorw should not be needed for
    add_fromtable().

    This tests fails in pyscal v0.7.7"""
    dframe = pd.DataFrame(
        columns=["SW", "KRW", "KROW", "PC"],
        data=[[sw, sw, 1 - sw, 1 - sw] for sw in [0, sw_mid, 1.0]],
    )
    wateroil = WaterOil(h=h, swl=0)
    wateroil.add_fromtable(dframe)
    assert wateroil.selfcheck()

    # GasOil did not fail in v0.7.7, but test anyway:
    gasoil = GasOil(h=h, swl=0)
    gasoil.add_fromtable(
        dframe, sgcolname="SW", krgcolname="KRW", krogcolname="KROW", pccolname="PCOW"
    )
    assert gasoil.selfcheck()


def test_ieee_754():
    """Test difficult values from experienced bugs related to
    IEE754 representation errors"""
    wateroilgas = WaterOilGas(swl=0.18)
    # For this particular swl value, we get:
    # wateroilgas.gasoil.table.sg.max()
    # Out[18]: 0.8200000000000001  # = (1 - 0.18)

    # Can we then interpolate from a table that goes up to 0.82?
    sgof = pd.DataFrame(
        columns=["SG", "KRG", "KROG", "PCOG"], data=[[0, 0, 1, 0], [0.82, 1, 0, 0]]
    )
    # Note, replacing 0.82 in the table above with 1-0.18 *is not the same*
    wateroilgas.gasoil.add_fromtable(sgof)
    assert wateroilgas.gasoil.table["KRG"].max() == 1.0
    assert wateroilgas.gasoil.table["KROG"].min() == 0.0
    assert wateroilgas.gasoil.table["PC"].min() == 0.0
    assert wateroilgas.gasoil.table["PC"].max() == 0.0


def test_wo_invalidcurves():
    """Test what happens when we give in invalid data"""
    # Sw data not ordered:
    krw1 = pd.DataFrame(columns=["SW", "KRW"], data=[[0.15, 0], [0.1, 1], [1, 1]])
    wateroil = WaterOil(swl=krw1["SW"].min(), h=0.1)
    with pytest.raises(ValueError):
        # pchip-interpolator raises this error;
        # x coordinates are not in increasing order
        wateroil.add_fromtable(krw1, krwcolname="KRW")

    krw2 = pd.DataFrame(
        columns=["SW", "KRW"], data=[[0.15, 0], [0.4, 0.6], [0.6, 0.4], [1, 1]]
    )
    wateroil = WaterOil(swl=krw2["SW"].min(), h=0.1)
    with pytest.raises(ValueError):
        # Should get notified that krw is not monotonous
        wateroil.add_fromtable(krw2, krwcolname="KRW")
    krow2 = pd.DataFrame(
        columns=["SW", "KROW"], data=[[0.15, 1], [0.4, 0.4], [0.6, 0.6], [1, 0]]
    )
    wateroil = WaterOil(swl=krow2["SW"].min(), h=0.1)
    with pytest.raises(ValueError):
        # Should get notified that krow is not monotonous
        wateroil.add_fromtable(krow2, krowcolname="KROW")
    pc2 = pd.DataFrame(
        columns=["SW", "PC"], data=[[0.15, 1], [0.4, 0.4], [0.6, 0.6], [1, 0]]
    )
    wateroil = WaterOil(swl=pc2["SW"].min(), h=0.1)
    with pytest.raises(ValueError):
        # Should get notified that pc is not monotonous
        wateroil.add_fromtable(pc2, pccolname="PC")

    pc3 = pd.DataFrame(
        columns=["SW", "PC"],
        data=[[0, np.inf], [0.1, 1], [0.4, 0.4], [0.6, 0.2], [1, 0]],
    )
    wateroil = WaterOil(swl=pc3["SW"].min(), h=0.1)
    # Will get warning that infinite numbers are ignored.
    # In this case the extrapolation is quite bad.
    wateroil.add_fromtable(pc3, pccolname="PC")

    # But when we later set swl larger, then we should
    # not bother about the infinity:
    wateroil = WaterOil(swl=0.1, h=0.1)
    wateroil.add_fromtable(pc3, pccolname="PC")
    assert np.isclose(wateroil.table["PC"].max(), pc3.iloc[1:]["PC"].max())

    # Choosing endpoint slightly to the left of 0.1 incurs
    # extrapolation. A warning will be given
    wateroil = WaterOil(swl=0.05, h=0.1)
    wateroil.add_fromtable(pc3, pccolname="PC")
    # Inequality due to extrapolation:
    assert wateroil.table["PC"].max() > pc3.iloc[1:]["PC"].max()


@pytest.mark.parametrize(
    "dframe, match",
    [
        (pd.DataFrame(columns=["SG"], data=[[0.15], [1]]), "sg must start at zero"),
        (
            pd.DataFrame(columns=["SG", "KRG"], data=[[0.0, 0], [0.1, 0.4], [0.1, 1]]),
            # error emitted by PchipInterpolator, not pyscal:
            "`x` must be strictly increasing",
        ),
        (
            pd.DataFrame(
                columns=["SG", "KRG"], data=[[0.0, 0], [0.4, 0.6], [0.6, 0.4], [1, 1]]
            ),
            "krg not increasing",
        ),
        (
            pd.DataFrame(
                columns=["SG", "KROG"], data=[[0, 1], [0.4, 0.4], [0.6, 0.6], [1, 0]]
            ),
            "krog not decreasing",
        ),
        (
            pd.DataFrame(
                columns=["SG", "PCOG"], data=[[0, 1], [0.4, 0.4], [0.6, 0.6], [1, 0]]
            ),
            "Incoming pc not increasing",
        ),
        (
            pd.DataFrame(
                columns=["SG", "KRG"], data=[[0, 1], [0.4, 0.6], [0.6, 0.4], [1, 1]]
            ),
            "Incoming krg not increasing",
        ),
        (
            pd.DataFrame(
                columns=["SG", "KROG"], data=[[0, 1.2], [0.4, 0.4], [0.6, 0.2], [1, 0]]
            ),
            "krog is above 1",
        ),
    ],
)
def test_go_invalidcurves(dframe, match):
    gasoil = GasOil(h=0.1)
    with pytest.raises(ValueError, match=match):
        gasoil.add_fromtable(dframe)


def test_wo_fromtable_problems():
    """Test wateroil from tables with problematic data"""
    # With default object:
    df1 = pd.DataFrame(
        columns=["SW", "KRW", "KROW", "PCOW"],
        data=[[0.15, 0, 1, 3], [0.89, 1, 0, 0.1], [1, 1, 0, 0]],
    )
    # With default object:
    wateroil = WaterOil(h=0.1)
    with pytest.raises(ValueError):
        wateroil.add_fromtable(df1)
        # This results in krw and krow overshooting 0 and 1
    # Fix left endpoint:
    wateroil = WaterOil(h=0.1, swl=df1["SW"].min())
    wateroil.add_fromtable(df1)
    # The table is now valid, but we did not preserve the 0.89 point
    check_table(wateroil.table)

    # If we also tell the WaterOil object about sorw, we are guaranteed
    # to have it expclitly included:
    wateroil = WaterOil(h=0.1, swl=df1["SW"].min(), sorw=1 - 0.89)
    wateroil.add_fromtable(df1)
    check_table(wateroil.table)
    # For low enough h, this will however NOT matter.


@pytest.mark.parametrize(
    "dframe, swl, exception, message",
    [
        (
            pd.DataFrame(
                columns=["SW", "KRW", "KROW", "PCOW"],
                data=[[0.15, "fooo", 1, 3], [0.89, 1, 0, 0.1], [1, 1, 0, 0]],
            ),
            0.15,
            ValueError,
            "Failed to parse column KRW",
        ),
        (
            pd.DataFrame(
                columns=["SW", "KRW", "KROW", "PCOW"],
                data=[[0.15, 1.1, 1, 3], [0.89, 1, 0, 0.1], [1, 1, 0, 0]],
            ),
            0.15,
            ValueError,
            "KRW is above 1",
        ),
        (
            pd.DataFrame(
                columns=["SW", "KRW", "KROW", "PCOW"],
                data=[[0.15, -0.1, 1, 3], [0.89, 1, 0, 0.1], [1, 1, 0, 0]],
            ),
            0.15,
            ValueError,
            "KRW is below 0",
        ),
        (
            pd.DataFrame(
                columns=["SW", "KRW", "KROW", "PCOW"],
                data=[[0.15, 0.1, "foo", 3], [0.89, 1, 0, 0.1], [1, 1, 0, 0]],
            ),
            0.15,
            ValueError,
            "Failed to parse column KROW",
        ),
        (
            pd.DataFrame(
                columns=["SW", "KRW", "KROW", "PCOW"],
                data=[[0.15, 0.1, 1.1, 3], [0.89, 1, 0, 0.1], [1, 1, 0, 0]],
            ),
            0.15,
            ValueError,
            "KROW is above 1",
        ),
        (
            pd.DataFrame(
                columns=["SW", "KRW", "KROW", "PCOW"],
                data=[[0.15, 0.1, 1, 3], [0.89, 1, 0, 0.1], [1, 1, -0.001, 0]],
            ),
            0.15,
            ValueError,
            "KROW is below 0",
        ),
        (
            pd.DataFrame(
                columns=["SW", "KRW", "KROW", "PCOW"],
                data=[[0.15, 0.1, 1, 3], [0.89, 1, 0, 0.1], [1, 1, -0.001, 0]],
            ),
            0.10,
            ValueError,
            "Incompatible swl",
        ),
        (
            pd.DataFrame(
                columns=["SW", "KRW", "KROW", "PCOW"],
                data=[[0.15, 0.1, 1, 3], [0.89, 1, 0, 0.1], [1, 1, -0.001, 0]],
            ),
            0,
            ValueError,
            "Incompatible swl",
        ),
        (
            pd.DataFrame(
                columns=["SW", "KRW", "KROW", "PCOW"],
                data=[[0.15, 0.1, 1, 3], [0.89, 1, 0, 0.1], [1, 1, -0.001, 0]],
            ),
            0.16,
            ValueError,
            "Incompatible swl",
        ),
    ],
)
def test_wo_from_table_exceptions(dframe, swl, exception, message):
    """Test exceptions when loading wateroil data from tables"""
    wateroil = WaterOil(h=0.1, swl=swl)
    with pytest.raises(exception, match=message):
        wateroil.add_fromtable(dframe)


def test_fromtable_types():
    """Test loading from a table with incorrect types"""

    # This frame is valid, but the type was wrong. This
    # can happen if data is via CSV files, and some other rows
    # ruin the numerical interpretation of a column.
    df1 = pd.DataFrame(
        columns=["SW", "KRW", "KROW", "PC"],
        data=[["0", "0", "1", "0"], ["1", "1", "0", "0"]],
    )
    wateroil = WaterOil(h=0.1)
    wateroil.add_fromtable(
        df1, swcolname="SW", krwcolname="KRW", krowcolname="KROW", pccolname="PC"
    )
    assert "KRW" in wateroil.table.columns
    assert "KROW" in wateroil.table.columns
    assert "PC" in wateroil.table.columns
    check_table(wateroil.table)

    gasoil = GasOil(h=0.1)
    gasoil.add_fromtable(
        df1, sgcolname="SW", krgcolname="KRW", krogcolname="KROW", pccolname="PC"
    )
    assert "KRG" in gasoil.table.columns
    assert "KROG" in gasoil.table.columns
    assert "PC" in gasoil.table.columns
    check_table(gasoil.table)

    # But this should not make sense.
    df2 = pd.DataFrame(
        columns=["SW", "KRW", "KROW", "PC"],
        data=[["0", {"foo": "bar"}, "1", "2"], ["1", "1", "0", "0"]],
    )
    wateroil = WaterOil(h=0.1)
    with pytest.raises((ValueError, TypeError)):
        wateroil.add_fromtable(
            df2, swcolname="SW", krwcolname="KRW", krowcolname="KROW", pccolname="PC"
        )
    gasoil = GasOil(h=0.1)
    with pytest.raises((ValueError, TypeError)):
        gasoil.add_fromtable(
            df2, sgcolname="SW", krgcolname="KRW", krogcolname="KROW", pccolname="PC"
        )


@given(st.floats(min_value=1e-5, max_value=1))
def test_wo_fromtable_h(h):
    """Test making curves from tabular data with random stepsize h"""
    df1 = pd.DataFrame(
        columns=["SW", "KRW", "KROW", "PCOW"],
        data=[[0.15, 0, 1, 3], [0.89, 1, 0, 0.1], [1, 1, 0, 0]],
    )
    wateroil = WaterOil(h=h, swl=0.15, sorw=1 - 0.89)
    wateroil.add_fromtable(df1)
    check_table(wateroil.table)



"""Test the PyscalFactory module"""

import os
from pathlib import Path

import numpy as np
import pandas as pd
import pytest

from pyscal import (
    GasOil,
    GasWater,
    PyscalFactory,  # PyscalFactory static functions are deprecated
    SCALrecommendation,
    WaterOil,
    WaterOilGas,
)
from pyscal.factory import (
    create_gas_oil,
    create_gas_water,
    create_pyscal_list,
    create_scal_recommendation,
    create_scal_recommendation_list,
    create_water_oil,
    create_water_oil_gas,
    infer_tabular_file_format,
    load_relperm_df,
    slicedict,
    sufficient_gas_oil_params,
    sufficient_gas_water_params,
    sufficient_water_oil_params,
)
from pyscal.utils.testing import check_table, sat_table_str_ok


def test_factory_wateroil():
    """Test that we can create curves from dictionaries of parameters"""
    # Factory refuses to create incomplete defaulted objects.
    with pytest.raises(ValueError):
        create_water_oil()

    with pytest.raises(TypeError):
        # (it must be a dictionary)
        create_water_oil(swirr=0.01)

    with pytest.raises(TypeError):
        create_water_oil(params="swirr 0.01")

    wateroil = create_water_oil(
        {
            "swirr": 0.01,
            "swl": 0.1,
            "bogus": "foobar",
            "tag": "Good sand",
            "nw": 3,
            "now": 2,
            "krwend": 0.2,
            "krwmax": 0.5,
        }
    )
    assert isinstance(wateroil, WaterOil)
    assert wateroil.swirr == 0.01
    assert wateroil.swl == 0.1
    assert wateroil.tag == "Good sand"
    assert "KRW" in wateroil.table
    assert "Corey" in wateroil.krwcomment
    assert wateroil.table["KRW"].max() == 0.2  # Because sorw==0 by default
    check_table(wateroil.table)
    sat_table_str_ok(wateroil.SWOF())
    sat_table_str_ok(wateroil.SWFN())

    wateroil = create_water_oil(
        {"nw": 3, "now": 2, "sorw": 0.1, "krwend": 0.2, "krwmax": 0.5}
    )
    assert isinstance(wateroil, WaterOil)
    assert "KRW" in wateroil.table
    assert "Corey" in wateroil.krwcomment
    assert wateroil.table["KRW"].max() == 0.5
    check_table(wateroil.table)
    sat_table_str_ok(wateroil.SWOF())
    sat_table_str_ok(wateroil.SWFN())

    # Ambiguous works, but we don't guarantee that this results
    # in LET or Corey.
    wateroil = create_water_oil({"nw": 3, "Lw": 2, "Ew": 2, "Tw": 2, "now": 3})
    assert "KRW" in wateroil.table
    assert "Corey" in wateroil.krwcomment or "LET" in wateroil.krwcomment
    check_table(wateroil.table)
    sat_table_str_ok(wateroil.SWOF())
    sat_table_str_ok(wateroil.SWFN())

    # Mixing Corey and LET
    wateroil = create_water_oil({"Lw": 2, "Ew": 2, "Tw": 2, "krwend": 1, "now": 4})
    assert isinstance(wateroil, WaterOil)
    assert "KRW" in wateroil.table
    assert wateroil.table["KRW"].max() == 1.0
    assert "LET" in wateroil.krwcomment
    check_table(wateroil.table)
    sat_table_str_ok(wateroil.SWOF())
    sat_table_str_ok(wateroil.SWFN())

    wateroil = create_water_oil(
        {"Lw": 2, "Ew": 2, "Tw": 2, "Low": 3, "Eow": 3, "Tow": 3, "krwend": 0.5}
    )
    assert isinstance(wateroil, WaterOil)
    assert "KRW" in wateroil.table
    assert "KROW" in wateroil.table
    assert wateroil.table["KRW"].max() == 0.5
    assert wateroil.table["KROW"].max() == 1
    assert "LET" in wateroil.krwcomment
    assert "LET" in wateroil.krowcomment
    check_table(wateroil.table)
    sat_table_str_ok(wateroil.SWOF())
    sat_table_str_ok(wateroil.SWFN())

    # Add capillary pressure
    wateroil = create_water_oil(
        {
            "swl": 0.1,
            "nw": 1,
            "now": 1,
            "a": 2,
            "b": -1,
            "poro_ref": 0.2,
            "perm_ref": 100,
            "drho": 200,
        }
    )
    assert "PC" in wateroil.table
    assert wateroil.table["PC"].max() > 0.0
    assert "Simplified J" in wateroil.pccomment
    check_table(wateroil.table)
    sat_table_str_ok(wateroil.SWOF())
    sat_table_str_ok(wateroil.SWFN())

    # Test that the optional gravity g is picked up:
    wateroil = create_water_oil(
        {
            "swl": 0.1,
            "nw": 1,
            "now": 1,
            "a": 2,
            "b": -1,
            "poro_ref": 0.2,
            "perm_ref": 100,
            "drho": 200,
            "g": 0,
        }
    )
    assert "PC" in wateroil.table
    assert wateroil.table["PC"].max() == 0.0
    check_table(wateroil.table)
    sat_table_str_ok(wateroil.SWOF())
    sat_table_str_ok(wateroil.SWFN())

    # Test petrophysical simple J:
    wateroil = create_water_oil(
        {
            "swl": 0.1,
            "nw": 1,
            "now": 1,
            "a_petro": 2,
            "b_petro": -1,
            "poro_ref": 0.2,
            "perm_ref": 100,
            "drho": 200,
        }
    )
    assert "PC" in wateroil.table
    assert wateroil.table["PC"].max() > 0.0
    assert "etrophysic" in wateroil.pccomment
    check_table(wateroil.table)
    sat_table_str_ok(wateroil.SWOF())
    sat_table_str_ok(wateroil.SWFN())

    # One pc param missing:
    wateroil = create_water_oil(
        {
            "swl": 0.1,
            "nw": 1,
            "now": 1,
            "a": 2,
            "b": -1,
            "perm_ref": 100,
            "drho": 200,
            "g": 0,
        }
    )
    assert "PC" not in wateroil.table


def test_fast_mode():
    """Test that the fast-flag is passed on to constructed objects

    Each object's own test code tests the actual effects of the fast flag"""
    wateroil = create_water_oil({"nw": 2, "now": 2})
    assert not wateroil.fast
    wateroil = create_water_oil({"nw": 2, "now": 2}, fast=True)
    assert wateroil.fast

    gasoil = create_gas_oil({"ng": 2, "nog": 2})
    assert not gasoil.fast
    gasoil = create_gas_oil({"ng": 2, "nog": 2}, fast=True)
    assert gasoil.fast

    gaswater = create_gas_water({"nw": 2, "ng": 2})
    assert not gaswater.gasoil.fast
    assert not gaswater.wateroil.fast
    gaswater = create_gas_water({"nw": 2, "ng": 2}, fast=True)
    assert gaswater.gasoil.fast
    assert gaswater.wateroil.fast
    assert gaswater.fast

    wateroilgas = create_water_oil_gas(
        {"nw": 2, "now": 2, "ng": 2, "nog": 2}, fast=True
    )
    assert wateroilgas.fast
    assert wateroilgas.wateroil.fast
    assert wateroilgas.gasoil.fast

    scalrec = create_scal_recommendation(
        {
            "low": {"nw": 2, "now": 2, "ng": 2, "nog": 2},
            "base": {"nw": 2, "now": 2, "ng": 2, "nog": 2},
            "high": {"nw": 2, "now": 2, "ng": 2, "nog": 2},
        },
        fast=True,
    )
    assert scalrec.low.fast
    assert scalrec.base.fast
    assert scalrec.high.fast

    interpolant = scalrec.interpolate(-0.5)
    assert interpolant.fast


def test_init_with_swlheight():
    """With sufficient parameters, swl will be calculated on the fly
    when initializing the WaterOil object"""
    wateroil = create_water_oil(
        {
            "swlheight": 200,
            "nw": 1,
            "now": 1,
            "swirr": 0.01,
            "a": 1,
            "b": -2,
            "poro_ref": 0.2,
            "perm_ref": 100,
            "drho": 200,
        }
    )
    assert np.isclose(wateroil.swl, 0.02480395)
    assert "swl=0.024" in wateroil.SWOF()

    with pytest.raises(
        ValueError,
        match="Can't initialize from SWLHEIGHT without sufficient simple-J parameters",
    ):
        # This should fail because capillary pressure parameters are not provided.
        create_water_oil({"swlheight": 200, "nw": 1, "now": 1})

    # swcr must be larger than swl:
    with pytest.raises(ValueError, match="lower than computed swl"):
        create_water_oil(
            {
                "swlheight": 200,
                "nw": 1,
                "now": 1,
                "swirr": 0.01,
                "swcr": 0.0101,
                "a": 1,
                "b": -2,
                "poro_ref": 0.2,
                "perm_ref": 100,
                "drho": 200,
            }
        )

    # swlheight must be positive:
    with pytest.raises(ValueError, match="swlheight must be larger than zero"):
        create_water_oil(
            {
                "swlheight": -200,
                "nw": 1,
                "now": 1,
                "swirr": 0.01,
                "a": 1,
                "b": -2,
                "poro_ref": 0.2,
                "perm_ref": 100,
                "drho": 200,
            }
        )

    # If swcr is large enough, it will pass:
    wateroil = create_water_oil(
        {
            "swlheight": 200,
            "nw": 1,
            "now": 1,
            "swirr": 0.01,
            "swcr": 0.3,
            "a": 1,
            "b": -2,
            "poro_ref": 0.2,
            "perm_ref": 100,
            "drho": 200,
        }
    )
    assert wateroil.swcr > wateroil.swl
    assert wateroil.swcr == 0.3
    assert "swcr=0.3" in wateroil.SWOF()

    # Test that GasWater also can be initialized with swlheight:
    gaswater = create_gas_water(
        {
            "swlheight": 200,
            "nw": 1,
            "ng": 1,
            "swirr": 0.01,
            "swcr": 0.3,
            "a": 1,
            "b": -2,
            "poro_ref": 0.2,
            "perm_ref": 100,
            "drho": 200,
        }
    )
    assert "swl=0.024" in gaswater.SWFN()
    assert gaswater.swcr > gaswater.swl
    assert gaswater.swcr == 0.3
    assert "swcr=0.3" in gaswater.SWFN()

    # Test error message for missing swirr when swlheight is asked for:
    with pytest.raises(
        ValueError, match="Can't initialize from SWLHEIGHT without sufficient simple-J"
    ):
        create_water_oil(
            {
                "swlheight": 200,
                "nw": 1,
                "now": 1,
                "a": 1,
                "b": -2,
                "poro_ref": 0.2,
                "perm_ref": 100,
                "drho": 200,
            }
        )


def test_relative_swcr():
    """swcr can be initialized relative to swl

    Relevant when swl is initialized from swlheight."""

    with pytest.raises(ValueError, match="swl must be provided"):
        create_water_oil({"swcr_add": 0.1, "nw": 1, "now": 1, "swirr": 0.01})
    with pytest.raises(ValueError, match="swcr and swcr_add at the same time"):
        create_water_oil(
            {"swcr_add": 0.1, "swcr": 0.1, "swl": 0.1, "nw": 1, "now": 1, "swirr": 0.01}
        )
    wateroil = create_water_oil(
        {"swcr_add": 0.1, "swl": 0.1, "nw": 1, "now": 1, "swirr": 0.01}
    )
    assert wateroil.swcr == 0.2

    # Test when relative to swlheight:
    wateroil = create_water_oil(
        {
            "swlheight": 200,
            "swcr_add": 0.01,
            "nw": 1,
            "now": 1,
            "swirr": 0.01,
            "a": 1,
            "b": -2,
            "poro_ref": 0.2,
            "perm_ref": 100,
            "drho": 200,
        }
    )
    assert np.isclose(wateroil.swl, 0.02480395)
    assert np.isclose(wateroil.swcr, 0.02480395 + 0.01)

    gaswater = create_gas_water(
        {
            "swlheight": 200,
            "nw": 1,
            "ng": 1,
            "swirr": 0.01,
            "swcr_add": 0.1,
            "a": 1,
            "b": -2,
            "poro_ref": 0.2,
            "perm_ref": 100,
            "drho": 200,
        }
    )
    assert np.isclose(gaswater.swl, 0.02480395)
    assert np.isclose(gaswater.swcr, 0.02480395 + 0.1)


def test_ambiguity():
    """Test how the factory handles ambiguity between Corey and LET
    parameters"""
    wateroil = create_water_oil(
        {"swl": 0.1, "nw": 10, "Lw": 1, "Ew": 1, "Tw": 1, "now": 2, "h": 0.1, "no": 2}
    )
    # Corey is picked here.
    assert "Corey" in wateroil.krwcomment
    assert "KRW" in wateroil.table


def test_factory_gasoil():
    """Test that we can create curves from dictionaries of parameters"""

    # Factory refuses to create incomplete defaulted objects.
    with pytest.raises(ValueError):
        create_gas_oil()

    with pytest.raises(TypeError):
        # (this must be a dictionary)
        create_gas_oil(swirr=0.01)

    with pytest.raises(TypeError):
        create_gas_oil(params="swirr 0.01")

    gasoil = create_gas_oil(
        {"swirr": 0.01, "swl": 0.1, "sgcr": 0.05, "tag": "Good sand", "ng": 1, "nog": 2}
    )
    assert isinstance(gasoil, GasOil)
    assert gasoil.sgcr == 0.05
    assert gasoil.sgro == 0.0
    assert gasoil.swl == 0.1
    assert gasoil.swirr == 0.01
    assert gasoil.tag == "Good sand"
    sgof = gasoil.SGOF()
    sat_table_str_ok(sgof)
    check_table(gasoil.table)
    assert "Corey krg" in sgof
    assert "Corey krog" in sgof
    assert "Zero capillary pressure" in sgof

    gasoil = create_gas_oil(
        {"ng": 1.2, "nog": 2, "krgend": 0.8, "krgmax": 0.9, "kroend": 0.6}
    )
    sgof = gasoil.SGOF()
    sat_table_str_ok(sgof)
    assert "kroend=0.6" in sgof
    assert "krgend=0.8" in sgof
    check_table(gasoil.table)

    gasoil = create_gas_oil({"ng": 1.3, "Log": 2, "Eog": 2, "Tog": 2})
    sgof = gasoil.SGOF()
    check_table(gasoil.table)
    sat_table_str_ok(sgof)
    assert "Corey krg" in sgof
    assert "LET krog" in sgof

    gasoil = create_gas_oil({"Lg": 1, "Eg": 1, "Tg": 1, "Log": 2, "Eog": 2, "Tog": 2})
    sgof = gasoil.SGOF()
    sat_table_str_ok(sgof)
    check_table(gasoil.table)
    assert "LET krg" in sgof
    assert "LET krog" in sgof


def test_factory_wog_gascondensate():
    """Test modelling of gas condensate, which in pyscal terms
    is the same as wateroilgas, except that we allow for aliasing
    in sgrw=sorw for the underlying WaterOil object, and also there
    are additional parameters sgro and kromax for GasOil."""
    wcg = create_water_oil_gas(
        {
            "nw": 2,
            "now": 3,
            "ng": 1,
            "nog": 2,
            "sgrw": 0.1,
            "swl": 0.1,
            "sgcr": 0.1,
            "sgro": 0.1,
            "kroend": 0.5,
            "kromax": 0.9,
        }
    )
    assert wcg.gasoil.sgro == 0.1
    assert wcg.wateroil.sorw == 0.1

    swof = wcg.SWOF()
    sgof = wcg.SGOF()

    # sgrw has been aliased to sorw, but the WaterOil object does not know that:
    assert "sgrw" not in swof
    assert "sorw=0.1" in swof
    assert "sgro=0.1" in sgof
    assert "kroend=0.5" in sgof
    assert "kromax=0.9" in sgof

    sat_table_str_ok(swof)
    sat_table_str_ok(sgof)

    # Different sorw and sgrw is a hard error:
    with pytest.raises(ValueError, match="must equal"):
        create_water_oil_gas(
            {"nw": 2, "now": 3, "ng": 1, "nog": 2, "sorw": 0.2, "sgrw": 0.1, "swl": 0.1}
        )

    # But it will pass if they both are supplied but are equal:
    wcg_2 = create_water_oil_gas(
        {"nw": 2, "now": 3, "ng": 1, "nog": 2, "sorw": 0.2, "sgrw": 0.2, "swl": 0.1}
    )
    assert "sorw=0.2" in wcg_2.SWOF()

    # kroend higher than kromax is an error:
    with pytest.raises(AssertionError):
        create_water_oil_gas(
            {
                "nw": 2,
                "now": 3,
                "ng": 1,
                "nog": 2,
                "sgcr": 0.1,
                "sgro": 0.1,
                "kromax": 0.5,
                "kroend": 0.8,
                "swl": 0.1,
            }
        )


def test_factory_go_gascondensate():
    """In gas condensate problems, the sgro and kromax parameters are relevant"""
    gasoil = create_gas_oil(
        {
            "sgro": 0.1,
            "sgcr": 0.1,
            "tag": "Good sand",
            "ng": 1,
            "nog": 2,
            "kroend": 0.5,
            "kromax": 0.9,
        }
    )
    assert isinstance(gasoil, GasOil)
    assert gasoil.sgro == 0.1
    assert gasoil.tag == "Good sand"
    sgof = gasoil.SGOF()
    sat_table_str_ok(sgof)
    check_table(gasoil.table)
    assert "Corey krog" in sgof
    assert "kroend=0.5" in sgof
    assert "kromax=0.9" in sgof
    assert "sgro=0.1" in sgof


def test_factory_gaswater():
    """Test that we can create gas-water curves from dictionaries of parameters"""

    # Factory refuses to create incomplete defaulted objects.
    with pytest.raises(ValueError):
        create_gas_water()

    with pytest.raises(TypeError):
        create_gas_water(swirr=0.01)

    with pytest.raises(TypeError):
        # (it must be a dictionary)
        create_gas_water(params="swirr 0.01")

    gaswater = create_gas_water(
        {
            "swirr": 0.01,
            "swl": 0.03,
            "sgrw": 0.1,
            "sgcr": 0.15,
            "tag": "gassy sand",
            "ng": 2,
            "nw": 2,
        }
    )

    assert isinstance(gaswater, GasWater)

    assert gaswater.swirr == 0.01
    assert gaswater.swl == 0.03
    assert gaswater.sgrw == 0.1
    assert gaswater.sgcr == 0.15
    assert gaswater.tag == "gassy sand"

    sgfn = gaswater.SGFN()
    swfn = gaswater.SWFN()
    sat_table_str_ok(sgfn)
    sat_table_str_ok(swfn)
    check_table(gaswater.wateroil.table)
    check_table(gaswater.gasoil.table)

    assert "sgrw=0.1" in swfn
    assert "swirr=0.01" in sgfn
    assert "swirr=0.01" in swfn
    assert "sgrw=0.1" in swfn
    assert "sgcr=0.15" in sgfn
    assert "nw=2" in swfn
    assert "ng=2" in sgfn
    assert "gassy sand" in sgfn

    gaswater = create_gas_water({"lg": 1, "eg": 1, "tg": 1, "nw": 3})

    sgfn = gaswater.SGFN()
    swfn = gaswater.SWFN()
    sat_table_str_ok(sgfn)
    sat_table_str_ok(swfn)
    check_table(gaswater.wateroil.table)
    check_table(gaswater.gasoil.table)


def test_factory_wateroilgas():
    """Test creating discrete cases of WaterOilGas from factory"""

    # Factory refuses to create incomplete defaulted objects.
    with pytest.raises(ValueError):
        create_water_oil_gas()

    with pytest.raises(TypeError):
        # (this must be a dictionary)
        create_water_oil_gas(swirr=0.01)

    with pytest.raises(TypeError):
        create_water_oil_gas(params="swirr 0.01")

    wog = create_water_oil_gas({"nw": 2, "now": 3, "ng": 1, "nog": 2.5})
    swof = wog.SWOF()
    sgof = wog.SGOF()
    sat_table_str_ok(swof)  # sgof code works for swof also currently
    sat_table_str_ok(sgof)
    assert "Corey krg" in sgof
    assert "Corey krog" in sgof
    assert "Corey krw" in swof
    assert "Corey krow" in swof
    check_table(wog.gasoil.table)
    check_table(wog.wateroil.table)

    # Some users will mess up lower vs upper case:
    wog = create_water_oil_gas({"NW": 2, "NOW": 3, "NG": 1, "nog": 2.5})
    swof = wog.SWOF()
    sgof = wog.SGOF()
    sat_table_str_ok(swof)  # sgof code works for swof also currently
    sat_table_str_ok(sgof)
    assert "Corey krg" in sgof
    assert "Corey krog" in sgof
    assert "Corey krw" in swof
    assert "Corey krow" in swof

    # Mangling data
    wateroil = create_water_oil_gas({"nw": 2, "now": 3, "ng": 1})
    assert wateroil.gasoil is None


def test_factory_wateroilgas_deprecated_krowgend():
    """Using long-time deprecated krowend and krogend will fail"""
    with pytest.raises(ValueError):
        create_water_oil_gas(
            {"nw": 2, "now": 3, "ng": 1, "nog": 2.5, "krowend": 0.6, "krogend": 0.7}
        )


def test_factory_wateroilgas_wo():
    """Test making only wateroil through the wateroilgas factory"""
    wog = create_water_oil_gas(
        {"nw": 2, "now": 3, "kroend": 0.5, "sorw": 0.04, "swcr": 0.1}
    )
    swof = wog.SWOF()
    assert "Corey krw" in swof
    assert "KRW" in wog.wateroil.table
    sat_table_str_ok(swof)
    check_table(wog.wateroil.table)
    assert wog.gasoil is None

    wog.SGOF()


def test_factory_wateroil_paleooil(caplog):
    """Test making a WaterOil object with socr different from sorw."""
    sorw = 0.09
    wateroil = create_water_oil(
        {
            "nw": 2,
            "now": 3,
            "kroend": 0.5,
            "sorw": sorw,
            "socr": sorw + 0.01,
            "swcr": 0.1,
        }
    )
    swof = wateroil.SWOF()
    assert "Corey krw" in swof
    assert "socr=0.1" in swof
    sat_table_str_ok(swof)
    check_table(wateroil.table)

    # If socr is close to sorw, socr is reset to sorw.
    for socr in [sorw - 1e-9, sorw, sorw + 1e-9]:
        wo_socrignored = create_water_oil(
            {"nw": 2, "now": 3, "kroend": 0.5, "sorw": 0.09, "socr": socr, "swcr": 0.1}
        )
        swof = wo_socrignored.SWOF()
        assert "socr" not in swof  # socr is effectively ignored when = sorw.
        sat_table_str_ok(swof)
        if socr != sorw:
            # This warning should only occur when it seems like the user
            # has tried to explicitly set socr
            assert "socr was close to sorw, reset to sorw" in caplog.text

    with pytest.raises(ValueError, match="socr must be equal to or larger than sorw"):
        create_water_oil(
            {
                "nw": 2,
                "now": 3,
                "kroend": 0.5,
                "sorw": 0.09,
                "socr": 0.001,
                "swcr": 0.1,
                "h": 0.1,
            }
        )


def test_load_relperm_df(tmp_path, caplog):
    """Test loading of dataframes with validation from excel or from csv"""
    testdir = Path(__file__).absolute().parent

    scalfile_xls = testdir / "data/scal-pc-input-example.xlsx"

    scaldata = load_relperm_df(scalfile_xls)
    with pytest.raises(IOError):
        load_relperm_df("not-existing-file")

    with pytest.raises(ValueError, match="Non-existing sheet-name"):
        load_relperm_df(scalfile_xls, sheet_name="foo")

    assert "SATNUM" in scaldata
    assert "CASE" in scaldata
    assert not scaldata.empty

    os.chdir(tmp_path)
    scaldata.to_csv("scal-input.csv")
    scaldata_fromcsv = load_relperm_df("scal-input.csv")
    assert "CASE" in scaldata_fromcsv
    assert not scaldata_fromcsv.empty
    scaldata_fromdf = load_relperm_df(scaldata_fromcsv)
    assert "CASE" in scaldata_fromdf
    assert "SATNUM" in scaldata_fromdf
    assert len(scaldata_fromdf) == len(scaldata_fromcsv) == len(scaldata)

    scaldata_fromcsv = load_relperm_df("scal-input.csv", sheet_name="foo")
    assert "Sheet name only relevant for XLSX files, ignoring foo" in caplog.text

    with pytest.raises(ValueError, match="Unsupported argument"):
        load_relperm_df({"foo": 1})

    # Perturb the dataframe, this should trigger errors
    with pytest.raises(ValueError):
        load_relperm_df(scaldata.drop("SATNUM", axis="columns"))
    wrongsatnums = scaldata.copy()
    wrongsatnums["SATNUM"] = wrongsatnums["SATNUM"] * 2
    with pytest.raises(ValueError):
        load_relperm_df(wrongsatnums)
    wrongsatnums = scaldata.copy()
    wrongsatnums["SATNUM"] = wrongsatnums["SATNUM"].astype(int)
    wrongsatnums = wrongsatnums[wrongsatnums["SATNUM"] > 2]
    with pytest.raises(ValueError):
        load_relperm_df(wrongsatnums)
    wrongcases = scaldata.copy()
    wrongcases["CASE"] = wrongcases["CASE"] + "ffooo"
    with pytest.raises(ValueError):
        load_relperm_df(wrongcases)

    with pytest.raises(ValueError):
        load_relperm_df(scaldata.drop(["Lw", "Lg"], axis="columns"))

    # Insert a NaN, this replicates what happens if cells are merged
    mergedcase = scaldata.copy()
    mergedcase.loc[3, "SATNUM"] = np.nan
    with pytest.raises(ValueError):
        load_relperm_df(mergedcase)

    relpermfile_xls = testdir / "data/relperm-input-example.xlsx"
    relpermdata = load_relperm_df(relpermfile_xls)
    assert "TAG" in relpermdata
    assert "SATNUM" in relpermdata
    assert "satnum" not in relpermdata  # always converted to upper-case
    assert len(relpermdata) == 3
    swof_str = create_pyscal_list(relpermdata, h=0.2).SWOF()
    assert "Åre 1.8" in swof_str
    assert "SATNUM 2" in swof_str  # Autogenerated in SWOF, generated by factory
    assert "SATNUM 3" in swof_str
    assert "foobar" in swof_str  # Random string injected in xlsx.

    # Make a dummy text file
    Path("dummy.txt").write_text("foo\nbar, com", encoding="utf8")
    with pytest.raises(ValueError):
        load_relperm_df("dummy.txt")

    # Make an empty csv file
    Path("empty.csv").write_text("", encoding="utf8")
    with pytest.raises(ValueError, match="Impossible to infer file format"):
        load_relperm_df("empty.csv")

    with pytest.raises(ValueError, match="SATNUM must be present"):
        load_relperm_df(pd.DataFrame())

    # Merge tags and comments if both are supplied
    Path("tagandcomment.csv").write_text(
        "SATNUM,nw,now,tag,comment\n1,1,1,a-tag,a-comment", encoding="utf8"
    )
    tagandcomment_df = load_relperm_df("tagandcomment.csv")
    assert (
        tagandcomment_df["TAG"].to_numpy()[0]
        == "SATNUM 1 tag: a-tag; comment: a-comment"
    )

    # Missing SATNUMs:
    Path("wrongsatnum.csv").write_text("SATNUM,nw,now\n1,1,1\n3,1,1", encoding="utf8")
    with pytest.raises(ValueError, match="Missing SATNUMs?"):
        load_relperm_df("wrongsatnum.csv")

    # Missing SATNUMs, like merged cells:
    Path("mergedcells.csv").write_text(
        "CASE,SATNUM,nw,now\nlow,,1,1\nlow,1,2,2\nlow,,3,32", encoding="utf8"
    )
    with pytest.raises(ValueError, match="Found not-a-number"):
        load_relperm_df("mergedcells.csv")

    # Missing SATNUMs, like merged cells:
    Path("mergedcellscase.csv").write_text(
        "CASE,SATNUM,nw,now\n,1,1,1\nlow,1,2,2\n,1,3,32", encoding="utf8"
    )
    with pytest.raises(ValueError, match="Found not-a-number"):
        load_relperm_df("mergedcellscase.csv")

    # Not valid CSV file
    Path("notvalidcsv.csv").write_text("SATNUM;nw;now\n1;1;1", encoding="utf-8")
    with pytest.raises(TypeError, match="Supplied file is not a valid CSV file"):
        load_relperm_df("notvalidcsv.csv")


def test_many_nans():
    """Excel or oocalc sometimes saves a xlsx file that gives all NaN rows and
    all-NaN columns, maybe some column setting that triggers Pandas to load
    them as actual columns/rows.

    Ensure we handle extra Nans in both directions"""
    nanframe = pd.DataFrame(
        [
            {"SATNUM": 1, "nw": 2, "now": 2, "Unnamed: 15": np.nan},
            {"SATNUM": np.nan, "nw": np.nan, "now": np.nan, "Unnamed: 15": np.nan},
        ]
    )
    wateroil_list = create_pyscal_list(load_relperm_df(nanframe))
    assert len(wateroil_list) == 1
    sat_table_str_ok(wateroil_list.SWOF())


def test_xls_factory():
    """Test/demonstrate how to go from data in an excel row to pyscal objects

    This test function predates the load_relperm_df() function, but can
    still be in here.
    """
    testdir = Path(__file__).absolute().parent

    xlsxfile = testdir / "data/scal-pc-input-example.xlsx"

    scalinput = pd.read_excel(xlsxfile, engine="openpyxl").set_index(["SATNUM", "CASE"])

    for (satnum, _), params in scalinput.iterrows():
        assert satnum
        wog = create_water_oil_gas(params.to_dict())
        swof = wog.SWOF()
        assert "LET krw" in swof
        assert "LET krow" in swof
        assert "Simplified J" in swof
        sgof = wog.SGOF()
        sat_table_str_ok(sgof)
        assert "LET krg" in sgof
        assert "LET krog" in sgof


def test_create_scal_recommendation_list():
    """Test the factory methods for making scalrecommendation lists"""
    testdir = Path(__file__).absolute().parent
    scalfile_xls = testdir / "data/scal-pc-input-example.xlsx"
    scaldata = load_relperm_df(scalfile_xls)

    scalrec_list = create_scal_recommendation_list(scaldata)
    assert len(scalrec_list) == 3
    assert scalrec_list.pyscaltype == SCALrecommendation

    # Erroneous input:
    with pytest.raises(ValueError, match="Too many cases supplied for SATNUM 2"):
        create_scal_recommendation_list(
            pd.DataFrame(
                columns=["SATNUM", "CASE", "NW", "NOW"],
                data=[
                    [1, "low", 1, 1],
                    [1, "base", 2, 2],
                    [1, "high", 3, 3],
                    [2, "low", 1, 1],
                    [2, "nearlylow", 1.4, 1.2],
                    [2, "base", 2, 2],
                    [2, "high", 3, 3],
                ],
            )
        )
    with pytest.raises(ValueError, match="Too few cases supplied for SATNUM 2"):
        create_scal_recommendation_list(
            pd.DataFrame(
                columns=["SATNUM", "CASE", "NW", "NOW"],
                data=[
                    [1, "low", 1, 1],
                    [1, "base", 2, 2],
                    [1, "high", 3, 3],
                    [2, "low", 1, 1],
                    [2, "high", 3, 3],
                ],
            )
        )


def test_create_pyscal_list():
    """Test the factory methods for making pyscal lists"""
    testdir = Path(__file__).absolute().parent
    scalfile_xls = testdir / "data/scal-pc-input-example.xlsx"
    scaldata = load_relperm_df(scalfile_xls)
    basecasedata = scaldata[scaldata["CASE"] == "base"].reset_index()
    relpermlist = create_pyscal_list(basecasedata)
    assert len(relpermlist) == 3
    assert relpermlist.pyscaltype == WaterOilGas

    wo_list = create_pyscal_list(
        basecasedata.drop(["Lg", "Eg", "Tg", "Log", "Eog", "Tog"], axis="columns")
    )

    assert len(wo_list) == 3
    assert wo_list.pyscaltype == WaterOil

    go_list = create_pyscal_list(
        basecasedata.drop(["Lw", "Ew", "Tw", "Low", "Eow", "Tow"], axis="columns")
    )

    assert len(go_list) == 3
    assert go_list.pyscaltype == GasOil

    gw_list = create_pyscal_list(
        basecasedata.drop(["Low", "Eow", "Tow", "Log", "Eog", "Tog"], axis="columns")
    )

    assert len(gw_list) == 3
    assert gw_list.pyscaltype == GasWater

    with pytest.raises(
        ValueError, match="Could not determine two or three phase from parameters"
    ):
        create_pyscal_list(basecasedata.drop(["Ew", "Eg"], axis="columns"))


def test_scalrecommendation():
    """Testing making SCAL rec from dict of dict."""

    scal_input = {
        "low": {"nw": 2, "now": 4, "ng": 1, "nog": 2},
        "BASE": {"nw": 3, "NOW": 3, "ng": 1, "nog": 2},
        "high": {"nw": 4, "now": 2, "ng": 1, "nog": 3},
    }
    scal = create_scal_recommendation(scal_input)

    with pytest.raises(ValueError, match="Input must be a dict"):
        create_scal_recommendation("low")

    # (not supported yet to make WaterOil only..)
    interp = scal.interpolate(-0.5)
    sat_table_str_ok(interp.SWOF())
    sat_table_str_ok(interp.SGOF())
    sat_table_str_ok(interp.SLGOF())
    sat_table_str_ok(interp.SOF3())
    check_table(interp.wateroil.table)
    check_table(interp.gasoil.table)

    # Check that we error if any of the parameters above is missing:
    for case in ["low", "BASE", "high"]:
        copy1 = scal_input.copy()
        del copy1[case]
        with pytest.raises(ValueError):
            create_scal_recommendation(copy1)

    go_only = scal_input.copy()
    del go_only["low"]["now"]
    del go_only["low"]["nw"]
    gasoil = create_scal_recommendation(go_only)
    assert gasoil.low.wateroil is None
    assert gasoil.base.wateroil is not None
    assert gasoil.high.wateroil is not None
    # SCALrecommendation of gasoil only works as long as you
    # don't try to ask for water data:
    assert "SGFN" in gasoil.interpolate(-0.4).SGFN()
    assert "SWOF" not in gasoil.interpolate(-0.2).SWOF()

    basehigh = scal_input.copy()
    del basehigh["low"]
    with pytest.raises(ValueError, match='"low" case not supplied'):
        create_scal_recommendation(basehigh)

    baselow = scal_input.copy()
    del baselow["high"]
    with pytest.raises(ValueError, match='"high" case not supplied'):
        create_scal_recommendation(baselow)

    with pytest.raises(
        ValueError, match="All values in parameter dict must be dictionaries"
    ):
        create_scal_recommendation(
            {"low": [1, 2], "base": {"swl": 0.1}, "high": {"swl": 0.1}}
        )


def test_scalrecommendation_gaswater():
    """Testing making SCAL rec from dict of dict for gaswater input"""

    scal_input = {
        "low": {"nw": 2, "ng": 1},
        "BASE": {"nw": 3, "ng": 1},
        "high": {"nw": 4, "ng": 1},
    }
    scal = create_scal_recommendation(scal_input, h=0.2)
    interp = scal.interpolate(-0.5, h=0.2)
    sat_table_str_ok(interp.SWFN())
    sat_table_str_ok(interp.SGFN())
    check_table(interp.wateroil.table)
    check_table(interp.gasoil.table)


def test_xls_scalrecommendation():
    """Test making SCAL recommendations from xls data"""
    testdir = Path(__file__).absolute().parent

    xlsxfile = testdir / "data/scal-pc-input-example.xlsx"
    scalinput = pd.read_excel(xlsxfile, engine="openpyxl").set_index(["SATNUM", "CASE"])
    for satnum in scalinput.index.levels[0].to_numpy():
        dictofdict = scalinput.loc[satnum, :].to_dict(orient="index")
        scalrec = create_scal_recommendation(dictofdict)
        scalrec.interpolate(+0.5)


def test_no_gasoil():
    """The command client does not support two-phase gas-oil, because
    that is most likely an sign of a user input error.
    (misspelled other columns f.ex).

    Make sure we fail in that case."""
    dframe = pd.DataFrame(columns=["SATNUM", "NOW", "NG"], data=[[1, 2, 2]])
    with pytest.raises(ValueError):
        load_relperm_df(dframe)


def test_check_deprecated_krowgend():
    """Up until pyscal 0.5.x, krogend and krowend were parameters
    to the oil curve parametrization for WaterOil and GasOil. From
    pyscal 0.6.0, krogend and krowend are merged to kroend.
    After pyscal 0.8 presence of krogend and krowend is a ValueError
    """
    with pytest.raises(ValueError):
        create_water_oil({"swl": 0.1, "nw": 2, "now": 2, "krowend": 0.4})

    with pytest.raises(ValueError):
        create_gas_oil({"swl": 0.1, "ng": 2, "nog": 2, "krogend": 0.4})

    # If krogend and kroend are both present, krogend is to be silently ignored
    # (random columns are in general accepted and ignored by pyscal)

    gasoil = create_gas_oil(
        {"swl": 0.1, "ng": 2, "nog": 2, "krogend": 0.4, "kroend": 0.3}
    )
    assert gasoil.table["KROG"].max() == 0.3

    wateroil = create_water_oil(
        {"swl": 0.1, "nw": 2, "now": 2, "krowend": 0.4, "kroend": 0.3}
    )
    assert wateroil.table["KROW"].max() == 0.3


def parse_gensatfuncline(conf_line):
    """Utility function that emulates how gensatfunc could parse
    its configuration lines in a pyscalfactory compatible fashion

    Args:
        conf_line (str): gensatfunc config line
    Returns:
        dict
    """

    # This is how the config line should be interpreted in terms of
    # pyscal parameters. Note that we are case insensitive in the
    # factory class
    line_syntax = [
        "CMD",
        "Lw",
        "Ew",
        "Tw",
        "Lo",
        "Eo",
        "To",
        "Sorw",
        "Swl",
        "krwend",
        "steps",
        "perm",
        "poro",
        "a",
        "b",
        "sigma_costau",
    ]

    if len(conf_line.split()) > len(line_syntax):
        raise ValueError("Too many items on gensatfunc confline")

    params = {}
    for idx, value in enumerate(conf_line.split()):
        if idx > 0:  # Avoid the CMD
            params[line_syntax[idx]] = float(value)

    # The 'steps' is not supported in pyscal, convert it:
    if "steps" in params:
        params["h"] = 1.0 / params["steps"]

    if "krwend" not in params:  # Last mandatory item
        raise ValueError("Too few items on gensatfunc confline")

    return params


def test_gensatfunc():
    """Test how the external tool gen_satfunc could use
    the factory functionality"""

    # Example config line for gen_satfunc:
    conf_line_pc = "RELPERM 4 2 1 3 2 1 0.15 0.10 0.5 20 100 0.2 0.22 -0.5 30"

    wateroil = create_water_oil(parse_gensatfuncline(conf_line_pc))
    swof = wateroil.SWOF()
    assert "0.17580" in swof  # krw at sw=0.65
    assert "0.0127" in swof  # krow at sw=0.65
    assert "Capillary pressure from normalized J-function" in swof
    assert "2.0669" in swof  # pc at swl

    conf_line_min = "RELPERM 1 2 3 1 2 3 0.1 0.15 0.5 20"
    wateroil = create_water_oil(parse_gensatfuncline(conf_line_min))
    swof = wateroil.SWOF()
    assert "Zero capillary pressure" in swof

    conf_line_few = "RELPERM 1 2 3 1 2 3"
    with pytest.raises(ValueError):
        parse_gensatfuncline(conf_line_few)

    # sigma_costau is missing here:
    conf_line_almost_pc = "RELPERM 4 2 1 3 2 1 0.15 0.10 0.5 20 100 0.2 0.22 -0.5"
    wateroil = create_water_oil(parse_gensatfuncline(conf_line_almost_pc))
    swof = wateroil.SWOF()
    # The factory will not recognize the normalized J-function
    # when costau is missing. Any error message would be the responsibility
    # of the parser
    assert "Zero capillary pressure" in swof


def test_sufficient_params():
    """Test the utility functions to determine whether
    WaterOil and GasOil object have sufficient parameters"""

    assert sufficient_gas_oil_params({"ng": 0, "nog": 0})
    # If it looks like the user meant to create GasOil, but only provided
    # data for krg, then might error hard. If the user did not provide
    # any data for GasOil, then the code returns False
    with pytest.raises(ValueError):
        sufficient_gas_oil_params({"ng": 0}, failhard=True)
    assert not sufficient_gas_oil_params({"ng": 0}, failhard=False)
    assert not sufficient_gas_oil_params({})
    with pytest.raises(ValueError):
        sufficient_gas_oil_params({"lg": 0}, failhard=True)
    assert not sufficient_gas_oil_params({"lg": 0}, failhard=False)
    assert sufficient_gas_oil_params(
        {"lg": 0, "eg": 0, "Tg": 0, "log": 0, "eog": 0, "tog": 0}
    )

    assert sufficient_water_oil_params({"nw": 0, "now": 0})
    with pytest.raises(ValueError):
        sufficient_water_oil_params({"nw": 0}, failhard=True)
    assert not sufficient_water_oil_params({})
    with pytest.raises(ValueError):
        sufficient_water_oil_params({"lw": 0}, failhard=True)
    assert sufficient_water_oil_params(
        {"lw": 0, "ew": 0, "Tw": 0, "low": 0, "eow": 0, "tow": 0}
    )


def test_sufficient_params_gaswater():
    """Test that we can detect sufficient parameters
    for gas-water only"""
    assert sufficient_gas_water_params({"nw": 0, "ng": 0})
    assert not sufficient_gas_water_params({"nw": 0, "nog": 0})
    assert sufficient_gas_water_params(
        {"lw": 0, "ew": 0, "tw": 0, "lg": 0, "eg": 0, "tg": 0}
    )
    assert not sufficient_gas_water_params({"lw": 0})
    assert not sufficient_gas_water_params({"lw": 0, "lg": 0})
    assert not sufficient_gas_water_params({"lw": 0, "lg": 0})

    with pytest.raises(ValueError):
        sufficient_gas_water_params({"lw": 0}, failhard=True)
    with pytest.raises(ValueError):
        sufficient_gas_water_params({"nw": 3}, failhard=True)

    assert sufficient_gas_water_params({"lw": 0, "ew": 0, "tw": 0, "ng": 0})
    assert sufficient_gas_water_params({"lg": 0, "eg": 0, "tg": 0, "nw": 0})
    assert not sufficient_gas_water_params({"lg": 0, "eg": 0, "tg": 0, "ng": 0})


def test_case_aliasing():
    """Test that we can use aliases for the CASE column
    in SCAL recommendations"""
    dframe = pd.DataFrame(
        columns=["SATNUM", "CASE", "Nw", "Now", "Ng", "Nog"],
        data=[
            [1, "pess", 2, 2, 1, 1],
            [1, "base", 3, 1, 1, 1],
            [1, "opt", 3, 1, 1, 1],
        ],
    )
    relperm_data = load_relperm_df(dframe)
    create_scal_recommendation_list(relperm_data, h=0.2).interpolate(-0.4)
    dframe = pd.DataFrame(
        columns=["SATNUM", "CASE", "Nw", "Now", "Ng", "Nog"],
        data=[
            [1, "pessimistic", 2, 2, 1, 1],
            [1, "base", 3, 1, 1, 1],
            [1, "optiMISTIc", 3, 1, 1, 1],
        ],
    )
    relperm_data = load_relperm_df(dframe)
    create_scal_recommendation_list(relperm_data, h=0.2).interpolate(-0.4)

    with pytest.raises(ValueError):
        load_relperm_df(
            pd.DataFrame(
                columns=["SATNUM", "CASE", "Nw", "Now", "Ng", "Nog"],
                data=[
                    [1, "FOOBAR", 2, 2, 1, 1],
                    [1, "base", 3, 1, 1, 1],
                    [1, "optIMIstiC", 3, 1, 1, 1],
                ],
            )
        )

    # Ambigous data:
    with pytest.raises(ValueError):
        amb = load_relperm_df(
            pd.DataFrame(
                columns=["SATNUM", "CASE", "Nw", "Now", "Ng", "Nog"],
                data=[
                    [1, "low", 2, 2, 1, 1],
                    [1, "pess", 5, 5, 5, 5],
                    [1, "base", 3, 1, 1, 1],
                    [1, "optIMIstiC", 3, 1, 1, 1],
                ],
            )
        )
        create_scal_recommendation_list(amb)

    # Missing a case
    with pytest.raises(ValueError):
        load_relperm_df(
            pd.DataFrame(
                columns=["SATNUM", "CASE", "Nw", "Now", "Ng", "Nog"],
                data=[[1, "base", 3, 1, 1, 1], [1, "optIMIstiC", 3, 1, 1, 1]],
            )
        )
    # Missing a case
    with pytest.raises(ValueError):
        load_relperm_df(
            pd.DataFrame(
                columns=["SATNUM", "CASE", "Nw", "Now", "Ng", "Nog"],
                data=[[1, "base", 3, 1, 1, 1]],
            )
        )


def test_socr_via_dframe():
    """Test that the "socr" parameter is picked up from a dataframe/xlsx input"""
    p_list = create_pyscal_list(
        load_relperm_df(
            pd.DataFrame(
                columns=["SATNUM", "Nw", "Now", "socr"],
                data=[[1, 2, 2, 0.5]],
            )
        )
    )
    assert "socr=0.5" in p_list.SWOF()


def test_swirr_partially_missing(tmp_path):
    """Test that swirr can be present for only a subset of the rows,
    and interpreted as zero when not there."""
    dframe = pd.DataFrame(
        columns=[
            "SATNUM",
            "Nw",
            "Now",
            "swl",
            "swirr",
            "a",
            "b",
            "poro_ref",
            "perm_ref",
            "drho",
        ],
        data=[
            [1, 2, 2, 0.2, 0.1, 2, -2, 0.2, 100, 300],
            [2, 3, 3, 0.1, np.nan, np.nan, np.nan, np.nan, np.nan],
        ],
    )
    relperm_data = load_relperm_df(dframe)
    p_list = create_pyscal_list(relperm_data, h=0.2)
    assert "a=2, b=-2" in p_list[1].pccomment
    assert p_list[2].pccomment == ""

    os.chdir(tmp_path)
    dframe.to_excel("partial_pc.xlsx")
    relperm_data_via_xlsx = load_relperm_df("partial_pc.xlsx")
    p_list = create_pyscal_list(relperm_data_via_xlsx, h=0.2)
    assert "a=2, b=-2" in p_list[1].pccomment
    assert p_list[2].pccomment == ""


def test_corey_let_mix():
    """Test that we can supply a dataframe where some SATNUMs
    have Corey and others have LET"""
    dframe = pd.DataFrame(
        columns=["SATNUM", "Nw", "Now", "Lw", "Ew", "Tw", "Ng", "Nog"],
        data=[[1, 2, 2, np.nan, np.nan, np.nan, 1, 1], [2, np.nan, 3, 1, 1, 1, 2, 2]],
    )
    relperm_data = load_relperm_df(dframe)
    p_list = create_pyscal_list(relperm_data, h=0.2)
    swof1 = p_list.pyscal_list[0].SWOF()
    swof2 = p_list.pyscal_list[1].SWOF()
    assert "Corey krw" in swof1
    assert "Corey krow" in swof1
    assert "LET krw" in swof2
    assert "Corey krow" in swof2


def test_infer_tabular_file_format(tmp_path, caplog):
    """Test code that infers the fileformat of files with tabular data"""
    testdir = Path(__file__).absolute().parent
    assert (
        infer_tabular_file_format(testdir / "data/scal-pc-input-example.xlsx") == "xlsx"
    )
    assert (
        infer_tabular_file_format(str(testdir / "data/scal-pc-input-example.xlsx"))
        == "xlsx"
    )
    assert (
        infer_tabular_file_format(testdir / "data/scal-pc-input-example.xls") == "xls"
    )
    os.chdir(tmp_path)
    pd.DataFrame([{"SATNUM": 1, "NW": 2}]).to_csv("some.csv", index=False)
    assert infer_tabular_file_format("some.csv") == "csv"

    Path("empty.csv").write_text("", encoding="utf8")
    assert infer_tabular_file_format("empty.csv") == ""
    # Ensure Pandas's error message got through:
    assert "No columns to parse from file" in caplog.text

    # We don't want ISO-8859 files, ensure we fail
    norw_chars = "Dette,er,en,CSV,fil\nmed,iso-8859:,æ,ø,å"
    Path("iso8859.csv").write_bytes(norw_chars.encode("iso-8859-1"))
    assert infer_tabular_file_format("iso8859.csv") == ""
    # Providing an error that this error was due to ISO-8859 and
    # nothing else is deemed too hard.
    Path("utf8.csv").write_bytes(norw_chars.encode("utf-8"))
    assert infer_tabular_file_format("utf8.csv") == "csv"

    # Write some random bytes to a file, this should with very
    # little probability give a valid xlsx/xls/csv file.
    Path("wrong.csv").write_bytes(os.urandom(100))
    assert infer_tabular_file_format("wrong.csv") == ""


@pytest.mark.parametrize(
    "orig_dict, keylist, expected_dict",
    [
        ({}, [], {}),
        ({"foo": 1}, [], {}),
        ({"foo": 1}, ["fo"], {}),
        ({"foo": 1}, ["foo"], {"foo": 1}),
        ({}, ["foo"], {}),
    ],
)
def test_slicedict(orig_dict, keylist, expected_dict):
    """Test that dictionaries can be sliced for subsets"""
    assert slicedict(orig_dict, keylist) == expected_dict


def test_deprecated_functions():
    """Test the deprecated functions"""

    # Load file used for testing
    testdir = Path(__file__).absolute().parent
    scalfile_xls = testdir / "data/scal-pc-input-example.xlsx"

    with pytest.warns(DeprecationWarning):
        PyscalFactory.create_water_oil({"nw": 2, "now": 2})

    with pytest.warns(DeprecationWarning):
        PyscalFactory.create_gas_oil({"ng": 2, "nog": 2})

    with pytest.warns(DeprecationWarning):
        PyscalFactory.create_water_oil_gas(
            {
                "nw": 2,
                "now": 2,
                "ng": 2,
                "nog": 2,
            }
        )

    with pytest.warns(DeprecationWarning):
        PyscalFactory.create_gas_water({"nw": 2, "ng": 2})

    with pytest.warns(DeprecationWarning):
        PyscalFactory.create_scal_recommendation(
            {
                "low": {"nw": 2, "now": 2, "ng": 2, "nog": 2},
                "base": {"nw": 2, "now": 2, "ng": 2, "nog": 2},
                "high": {"nw": 2, "now": 2, "ng": 2, "nog": 2},
            }
        )

    with pytest.warns(DeprecationWarning):
        scaldata = PyscalFactory.load_relperm_df(scalfile_xls)

    with pytest.warns(DeprecationWarning):
        PyscalFactory.create_scal_recommendation_list(scaldata)

    with pytest.warns(DeprecationWarning):
        PyscalFactory.create_pyscal_list(scaldata)

    with pytest.warns(DeprecationWarning):
        PyscalFactory.remap_validate_cases(scaldata["CASE"].to_numpy())

    with pytest.warns(DeprecationWarning):
        PyscalFactory.create_wateroilgas_list(scaldata)

    with pytest.warns(DeprecationWarning):
        PyscalFactory.create_wateroil_list(scaldata)

    with pytest.warns(DeprecationWarning):
        PyscalFactory.create_gasoil_list(scaldata)

    with pytest.warns(DeprecationWarning):
        PyscalFactory.create_gaswater_list(scaldata)

    with pytest.warns(DeprecationWarning):
        PyscalFactory.alias_sgrw({"sgrw": 0.1})



"""Test the PyscalList module"""

from pathlib import Path

import numpy as np
import pandas as pd
import pytest

from pyscal import (
    GasOil,
    GasWater,
    PyscalList,
    SCALrecommendation,
    WaterOil,
    WaterOilGas,
)
from pyscal.factory import (
    create_pyscal_list,
    create_scal_recommendation_list,
    load_relperm_df,
)
from pyscal.utils.testing import sat_table_str_ok

try:
    import ecl2df

    HAVE_ECL2DF = True
except ImportError:
    HAVE_ECL2DF = False


def test_pyscallist_basic():
    """Test that the class acts like a list"""
    p_list = PyscalList()
    assert isinstance(p_list, PyscalList)
    assert not p_list

    p_list.append(None)
    assert not p_list
    p_list.append([])
    assert not p_list

    with pytest.raises(ValueError):
        p_list.append(1)
    with pytest.raises(ValueError):
        p_list.append("foo")

    p_list.append(WaterOil())
    assert len(p_list) == 1
    assert isinstance(p_list[1], WaterOil)
    with pytest.raises(IndexError):
        p_list[0]
    with pytest.raises(IndexError):
        p_list[2]
    with pytest.raises(ValueError):
        p_list.append(GasOil())
    assert len(p_list) == 1

    p_list.append(WaterOil())
    assert len(p_list) == 2

    p_list.append([WaterOil()])
    assert len(p_list) == 3

    with pytest.raises(ValueError, match="Not a pyscal object"):
        PyscalList([{}])

    init_from_list = PyscalList([WaterOil(), WaterOil()])
    assert len(init_from_list) == 2

    init_from_itself = PyscalList(PyscalList([WaterOil(), WaterOil()]))
    assert len(init_from_itself) == 2


def test_load_scalrec():
    """Load a SATNUM range from xlsx"""
    testdir = Path(__file__).absolute().parent

    scalrec_data = load_relperm_df(testdir / "data/scal-pc-input-example.xlsx")

    # Also check that we can read the old excel format
    scalrec_data_legacy_xls = load_relperm_df(
        testdir / "data/scal-pc-input-example.xls"
    )
    pd.testing.assert_frame_equal(scalrec_data, scalrec_data_legacy_xls)

    scalrec_list = create_scal_recommendation_list(scalrec_data)
    wog_list = scalrec_list.interpolate(-0.3)

    with pytest.raises((AssertionError, ValueError)):
        scalrec_list.interpolate(-1.1)
    assert wog_list.pyscaltype == WaterOilGas

    with pytest.raises(TypeError):
        scalrec_list.build_eclipse_data(family=1)
    with pytest.raises(TypeError):
        scalrec_list.build_eclipse_data(family=2)
    with pytest.raises(TypeError):
        scalrec_list.SWOF()

    scalrec_list.interpolate([-1, 1, 0])
    with pytest.raises((AssertionError, ValueError)):
        scalrec_list.interpolate([-1, 0, 1.1])

    with pytest.raises(
        TypeError, match="Can only interpolate PyscalList of type SCALrecommendation"
    ):
        wog_list.interpolate(-0.5)

    scalrec_list.interpolate(-1, 1)
    scalrec_list.interpolate(-1, [0, -1, 1])

    with pytest.raises(
        ValueError, match="Too few interpolation parameters given for WaterOil"
    ):
        scalrec_list.interpolate([-1, 1])
    with pytest.raises(
        ValueError, match="Too few interpolation parameters given for GasOil"
    ):
        scalrec_list.interpolate(1, [-1, 1])
    with pytest.raises(
        ValueError, match="Too many interpolation parameters given for WaterOil"
    ):
        scalrec_list.interpolate([-1, 1, 0, 0])
    with pytest.raises(
        ValueError, match="Too many interpolation parameters given for GasOil"
    ):
        scalrec_list.interpolate(1, [-1, 1, 0, 0])

    # Test slicing the scalrec to base, this is relevant for API usage.
    base_data = scalrec_data[scalrec_data["CASE"] == "base"].drop("CASE", axis=1)
    load_relperm_df(base_data)  # Ensure no errors.
    pyscal_list = create_pyscal_list(base_data)
    assert "LET" in pyscal_list.build_eclipse_data(family=1)


def test_df():
    """Test dataframe dumps"""
    testdir = Path(__file__).absolute().parent

    scalrec_data = load_relperm_df(testdir / "data/scal-pc-input-example.xlsx")
    scalrec_list = create_scal_recommendation_list(scalrec_data)
    wog_list = scalrec_list.interpolate(-0.3)

    # Test dataframe dumps:
    dframe = scalrec_list.df()
    assert "SG" in dframe
    assert "KRG" in dframe
    assert "KROG" in dframe
    assert "PCOW" in dframe
    if "PCOG" in dframe:
        # Allow PCOG to be included later.
        assert dframe["PCOG"].sum() == 0
    assert "KRW" in dframe
    assert "KROW" in dframe
    assert "SATNUM" in dframe
    assert dframe["SATNUM"].min() == 1
    assert len(dframe["SATNUM"].unique()) == len(scalrec_list)
    assert set(dframe["CASE"]) == {"pess", "base", "opt"}
    assert dframe["SATNUM"].max() == len(scalrec_list)
    if HAVE_ECL2DF:
        # Test using ecl2df to do the include file printing. First we need to
        # massage the dataframe into what ecl2df can handle:
        base_df_swof = (
            dframe.set_index("CASE")
            .loc["base"][["SW", "KRW", "KROW", "PCOW", "SATNUM"]]
            .assign(KEYWORD="SWOF")
            .dropna()
            .reset_index(drop=True)
        )
        ecl_inc = ecl2df.satfunc.df2ecl(base_df_swof)
        dframe_from_inc = ecl2df.satfunc.df(ecl_inc)
        pd.testing.assert_frame_equal(base_df_swof, dframe_from_inc)

        # Test also SGOF
        base_df_sgof = (
            dframe.set_index("CASE")
            .loc["base"][["SG", "KRG", "KROG", "SATNUM"]]
            .assign(KEYWORD="SGOF", PCOG=0.0)
            .dropna()
            .reset_index(drop=True)
        )
        ecl_inc = ecl2df.satfunc.df2ecl(base_df_sgof)
        dframe_from_inc = ecl2df.satfunc.df(ecl_inc)
        pd.testing.assert_frame_equal(base_df_sgof, dframe_from_inc, check_like=True)

    # WaterOilGasList:
    dframe = wog_list.df()
    assert "SG" in dframe
    assert "KRG" in dframe
    assert "KROG" in dframe
    assert "PCOW" in dframe
    assert "PCOG" in dframe  # This gets included through interpolation
    assert "KRW" in dframe
    assert "KROW" in dframe
    assert "SATNUM" in dframe
    assert dframe["SATNUM"].min() == 1
    assert len(dframe["SATNUM"].unique()) == len(wog_list)
    assert "CASE" not in dframe
    assert dframe["SATNUM"].max() == len(wog_list)

    # WaterOil list
    input_dframe = pd.DataFrame(columns=["SATNUM", "Nw", "Now"], data=[[1, 2, 2]])
    relperm_data = load_relperm_df(input_dframe)
    wo_list = create_pyscal_list(relperm_data, h=0.2)
    dframe = wo_list.df()
    assert "SW" in dframe
    assert "KRW" in dframe
    assert "KROW" in dframe
    assert "PCOW" not in dframe  # to be interpreted as zero
    assert "SATNUM" in dframe
    assert len(dframe.columns) == 4
    assert not dframe.empty

    # GasOil list
    input_dframe = pd.DataFrame(columns=["SATNUM", "Ng", "Nog"], data=[[1, 2, 2]])
    relperm_data = load_relperm_df(input_dframe)
    go_list = create_pyscal_list(relperm_data, h=0.2)
    dframe = go_list.df()
    assert "SG" in dframe
    assert "KRG" in dframe
    assert "KROG" in dframe
    assert "PCOG" not in dframe  # to be interpreted as zero
    assert "SATNUM" in dframe
    assert len(dframe.columns) == 4
    assert not dframe.empty

    gasoil_family1 = go_list.build_eclipse_data(family=1)
    assert "SGOF" in gasoil_family1

    with pytest.raises(ValueError, match="SLGOF not meaningful for GasOil"):
        go_list.build_eclipse_data(family=1, slgof=True)


def test_load_scalrec_tags():
    """Test tag handling for a SCAL recommendation with SATNUM range"""
    testdir = Path(__file__).absolute().parent

    scalrec_data = load_relperm_df(testdir / "data/scal-pc-input-example.xlsx")
    scalrec_list = create_scal_recommendation_list(scalrec_data)

    wog_list = scalrec_list.interpolate(-1)

    swof = wog_list.SWOF()
    assert swof.count("SCAL recommendation interpolation to -1") == 3
    assert swof.count("SATNUM 1") == 1
    assert swof.count("SATNUM") == 3

    sof3 = wog_list.SOF3()
    assert sof3.count("SCAL recommendation interpolation to -1") == 3
    assert sof3.count("SATNUM 1") == 1
    assert sof3.count("SATNUM") == 3

    assert (
        scalrec_list.interpolate(1)
        .SWOF()
        .count("SCAL recommendation interpolation to 1\n")
        == 3
    )
    assert (
        scalrec_list.interpolate(0)
        .SWOF()
        .count("SCAL recommendation interpolation to 0\n")
        == 3
    )
    assert (
        scalrec_list.interpolate(-0.444)
        .SWOF()
        .count("SCAL recommendation interpolation to -0.444\n")
        == 3
    )

    # Give each row in the SCAL recommendation its own tag (low, base, high
    # explicitly in each tag, someone will do that)
    scalrec_data["TAG"] = [
        "SAT1 low",
        "SAT1 base",
        "SAT1 high",
        "SAT2 pess",
        "SAT2 base",
        "SAT2 opt",
        "SAT3 pessimistic",
        "SAT3 base case",
        "SAT3 optimistic",
    ]
    scalrec_list2 = create_scal_recommendation_list(scalrec_data)
    swof = scalrec_list2.interpolate(-0.5, h=0.2).SWOF()
    assert swof.count("SCAL recommendation") == 3
    for tag in scalrec_data["TAG"]:
        assert swof.count(tag) == 1


def test_deprecated_dump_to_file(tmpdir):
    """Test dumping Eclipse include data to file. This
    functionality is deprecated in pyscallist"""
    testdir = Path(__file__).absolute().parent

    relperm_data = load_relperm_df(testdir / "data/relperm-input-example.xlsx")
    pyscal_list = create_pyscal_list(relperm_data)

    fam1 = pyscal_list.build_eclipse_data(family=1)
    sat_table_str_ok(fam1)

    fam2 = pyscal_list.build_eclipse_data(family=2)
    sat_table_str_ok(fam2)

    # Empty PyscalLists should return empty strings:
    assert PyscalList().build_eclipse_data(family=1) == ""
    assert PyscalList().build_eclipse_data(family=2) == ""

    tmpdir.chdir()
    Path("outputdir").mkdir(exist_ok=True)

    with pytest.warns(DeprecationWarning):
        pyscal_list.dump_family_1(filename="outputdir/output-fam1.inc")
    assert "SWOF" in Path("outputdir/output-fam1.inc").read_text(encoding="utf8")

    with pytest.warns(DeprecationWarning):
        pyscal_list.dump_family_2(filename="outputdir/output-fam2.inc")
    assert "SOF3" in Path("outputdir/output-fam2.inc").read_text(encoding="utf8")

    with pytest.warns(DeprecationWarning):
        pyscal_list.SWOF(write_to_filename="swof.inc")
    assert "SWOF" in Path("swof.inc").read_text(encoding="utf8")


def test_capillary_pressure():
    """Test that we recognize capillary pressure parametrizations"""
    dframe = pd.DataFrame(
        columns=[
            "SATNUM",
            "nw",
            "now",
            "swl",
            "a",
            "b",
            "PORO_reF",
            "PERM_ref",
            "drho",
        ],
        data=[[1, 1, 1, 0.05, 3.6, -3.5, 0.25, 15, 150]],
    )
    pyscal_list = create_pyscal_list(load_relperm_df(dframe))
    swof = pyscal_list.build_eclipse_data(family=1)
    assert "Simplified J-function" in swof
    assert "petrophysical" not in swof

    dframe = pd.DataFrame(
        columns=[
            "SATNUM",
            "nw",
            "now",
            "swl",
            "a_petro",
            "b_petro",
            "PORO_reF",
            "PERM_ref",
            "drho",
        ],
        data=[[1, 1, 1, 0.05, 3.6, -3.5, 0.25, 15, 150]],
    )
    pyscal_list = create_pyscal_list(load_relperm_df(dframe))
    swof = pyscal_list.build_eclipse_data(family=1)
    assert "Simplified J-function" in swof
    assert "petrophysical" in swof

    dframe = pd.DataFrame(
        columns=[
            "SATNUM",
            "nw",
            "now",
            "swl",
            "a",
            "b",
            "PORO",
            "PERM",
            "sigma_COStau",
        ],
        data=[[1, 1, 1, 0.05, 3.6, -3.5, 0.25, 15, 30]],
    )
    pyscal_list = create_pyscal_list(load_relperm_df(dframe))
    swof = pyscal_list.build_eclipse_data(family=1)
    assert "normalized J-function" in swof
    assert "sigma_costau" in swof
    assert "petrophysical" not in swof


def test_swl_from_height():
    """Test that can initialize swl from capillary pressure height"""
    df_columns = [
        "SATNUM",
        "nw",
        "now",
        "swl",
        "swlheight",
        "swirr",
        "a",
        "b",
        "PORO_reF",
        "PERM_ref",
        "drho",
    ]
    dframe = pd.DataFrame(
        columns=df_columns,
        data=[[1, 1, 1, np.nan, 300, 0.00, 3.6, -3.5, 0.25, 15, 150]],
    )
    pyscal_list = create_pyscal_list(load_relperm_df(dframe))

    # Mix swlheight init and direct swl-init:
    assert np.isclose(pyscal_list[1].swl, 0.157461)
    dframe = pd.DataFrame(
        columns=df_columns,
        data=[
            [1, 1, 1, np.nan, 300, 0.00, 3.6, -3.5, 0.25, 15, 150],
            [2, 1, 1, 0.3, np.nan, 0.00, 3.6, -3.5, 0.25, 15, 150],
        ],
    )
    pyscal_list = create_pyscal_list(load_relperm_df(dframe))
    assert np.isclose(pyscal_list[1].swl, 0.157461)
    assert np.isclose(pyscal_list[2].swl, 0.3)

    # Ambiguous, swl and swlheight both supplied:
    dframe = pd.DataFrame(
        columns=df_columns,
        data=[[1, 1, 1, 0.3, 300, 0.00, 3.6, -3.5, 0.25, 15, 150]],
    )
    with pytest.raises(ValueError):
        create_pyscal_list(load_relperm_df(dframe))

    # WaterOilGas (gasoil is also dependant on the computed swl)
    df_wog_columns = [
        "SATNUM",
        "nw",
        "now",
        "ng",
        "nog",
        "swlheight",
        "swirr",
        "a",
        "b",
        "PORO_reF",
        "PERM_ref",
        "drho",
    ]
    dframe = pd.DataFrame(
        columns=df_wog_columns,
        data=[[1, 1, 1, 2, 2, 300, 0.00, 3.6, -3.5, 0.25, 15, 150]],
    )
    pyscal_list = create_pyscal_list(load_relperm_df(dframe))
    assert np.isclose(pyscal_list[1].wateroil.swl, 0.157461)
    assert np.isclose(pyscal_list[1].gasoil.swl, 0.157461)

    # Test for GasWater:
    df_gw_columns = [
        "SATNUM",
        "nw",
        "ng",
        "swlheight",
        "swirr",
        "a",
        "b",
        "PORO_REF",
        "PERM_REF",
        "drho",
    ]
    dframe = pd.DataFrame(
        columns=df_gw_columns,
        data=[[1, 2, 3, 300, 0.00, 3.6, -3.5, 0.25, 15, 150]],
    )
    pyscal_list = create_pyscal_list(load_relperm_df(dframe))
    assert np.isclose(pyscal_list[1].wateroil.swl, 0.157461)
    assert np.isclose(pyscal_list[1].swl, 0.157461)


def test_twophase_scal_wateroil():
    """Test interpolation for a two-phase setup water-oil"""
    # Triggers bug in pyscal <= 0.5.0
    dframe = pd.DataFrame(
        columns=["SATNUM", "CASE", "NW", "NOW", "TAG"],
        data=[
            [1, "pess", 1, 1, "thetag"],
            [1, "base", 2, 2, "thetag"],
            [1, "opt", 3, 3, "thetag"],
        ],
    )
    pyscal_list = create_scal_recommendation_list(load_relperm_df(dframe))
    pyscal_list = pyscal_list.interpolate(-0.5)

    # This works, but provides an error message for Gas.
    # The correct usage would be to call .SWOF(), but the pyscal
    # client calls 1 on two-phase problems.
    swof = pyscal_list.build_eclipse_data(family=1)
    assert "thetag" in swof
    assert "SWOF" in swof
    assert "SGOF" not in swof


def test_twophase_scal_gasoil():
    """Test interpolation for a two-phase setup gas-oil"""
    dframe = pd.DataFrame(
        columns=["SATNUM", "CASE", "NG", "NOG", "TAG"],
        data=[
            [1, "pess", 1, 1, "thetag"],
            [1, "base", 2, 2, "thetag"],
            [1, "opt", 3, 3, "thetag"],
        ],
    )
    pyscal_list = create_scal_recommendation_list(load_relperm_df(dframe))
    pyscal_list = pyscal_list.interpolate(-0.5)

    sgof = pyscal_list.build_eclipse_data(family=1)
    assert "thetag" in sgof
    assert "SGOF" in sgof
    assert "SWOF" not in sgof


def test_gaswater():
    """Test list of gas-water objects"""
    dframe = pd.DataFrame(
        columns=["SATNUM", "NW", "NG", "TAG"],
        data=[[1, 2, 2, "thetag"], [2, 3, 3, "othertag"]],
    )
    pyscal_list = create_pyscal_list(load_relperm_df(dframe), h=0.1)
    assert pyscal_list.pyscaltype == GasWater
    dump = pyscal_list.build_eclipse_data(family=2)
    assert "SATNUM 2" in dump
    assert "SATNUM 1" in dump
    assert "SATNUM 3" not in dump
    assert "SWFN" in dump
    assert "SGFN" in dump
    assert "othertag" in dump
    assert "thetag" in dump

    with pytest.raises(ValueError, match="Family 1 output not possible for GasWater"):
        pyscal_list.build_eclipse_data(family=1)

    sgfn = pyscal_list.SGFN()
    assert "SGFN" in sgfn
    assert "krw = krg" in sgfn
    swfn = pyscal_list.SWFN()
    assert "SWFN" in swfn
    assert "krw = krg" in swfn

    with pytest.raises(
        AttributeError, match="'GasWater' object has no attribute 'SWOF'"
    ):
        pyscal_list.SWOF()


def test_gaswater_scal():
    """Test list of gas-water objects in scal recommendation"""
    dframe = pd.DataFrame(
        columns=["SATNUM", "CASE", "NW", "NG", "TAG"],
        data=[
            [1, "pess", 2, 2, "thetag"],
            [1, "base", 3, 3, "thetag"],
            [1, "opt", 4, 4, "thetag"],
        ],
    )
    pyscal_list = create_scal_recommendation_list(load_relperm_df(dframe), h=0.1)
    assert pyscal_list.pyscaltype == SCALrecommendation

    dump = pyscal_list.interpolate(-1).build_eclipse_data(family=2)
    assert "SWFN" in dump
    assert "SGFN" in dump


def test_explicit_df():
    """Test some dataframes, check the error messages given"""

    dframe = pd.DataFrame(columns=["satnum"], data=[[1], [2]])
    with pytest.raises(ValueError):
        # SATNUM column must be upper case, or should we allow lowercase??
        relperm_data = load_relperm_df(dframe)

    dframe = pd.DataFrame(columns=["SATNUM"], data=[[0], [1]])
    with pytest.raises(ValueError):
        # SATNUM must start at 1.
        relperm_data = load_relperm_df(dframe)

    dframe = pd.DataFrame(columns=["SATNUM"], data=[[1], ["foo"]])
    with pytest.raises(ValueError):
        # SATNUM must contain only integers
        relperm_data = load_relperm_df(dframe)

    dframe = pd.DataFrame(
        columns=["SATNUM", "nw", "now"], data=[[1.01, 1, 1], [2.01, 1, 1]]
    )
    # This one will pass, as these can be converted to ints
    relperm_data = load_relperm_df(dframe)

    dframe = pd.DataFrame(
        columns=["SATNUM", "nw", "now"], data=[[1.01, 1, 1], [1.3, 1, 1]]
    )
    with pytest.raises(ValueError):
        # complains about non-uniqueness in SATNUM
        relperm_data = load_relperm_df(dframe)

    dframe = pd.DataFrame(columns=["SATNUM"], data=[[1], [2]])
    with pytest.raises(ValueError):
        # Not enough data
        relperm_data = load_relperm_df(dframe)

    # Minimal amount of data:
    dframe = pd.DataFrame(columns=["SATNUM", "Nw", "Now"], data=[[1, 2, 2]])
    relperm_data = load_relperm_df(dframe)
    p_list = create_pyscal_list(relperm_data, h=0.2)
    p_list.build_eclipse_data(family=1)
    with pytest.raises(ValueError):
        p_list.build_eclipse_data(family=2)

    # Case insensitive for parameters
    dframe = pd.DataFrame(columns=["SATNUM", "nw", "now"], data=[[1, 2, 2]])
    relperm_data = load_relperm_df(dframe)
    p_list = create_pyscal_list(relperm_data, h=0.2)
    p_list.build_eclipse_data(family=1)

    # Minimal wateroilgas
    dframe = pd.DataFrame(
        columns=["SATNUM", "Nw", "Now", "ng", "nOG"], data=[[1, 2, 2, 2, 2]]
    )
    relperm_data = load_relperm_df(dframe)
    p_list = create_pyscal_list(relperm_data, h=0.2)
    relperm_str = p_list.build_eclipse_data(family=1)
    assert "SWOF" in relperm_str
    assert "SGOF" in relperm_str
    assert "SLGOF" not in relperm_str
    assert "SOF3" not in relperm_str

    # Minimal wateroilgas with pc
    dframe = pd.DataFrame(
        columns=[
            "SATNUM",
            "swl",
            "Nw",
            "Now",
            "ng",
            "nOG",
            "a",
            "b",
            "poro_ref",
            "perm_ref",
            "drho",
        ],
        data=[[1, 0.1, 2, 2, 2, 2, 1.5, -0.5, 0.1, 100, 300]],
    )
    relperm_data = load_relperm_df(dframe)
    p_list = create_pyscal_list(relperm_data, h=0.2)
    relperm_str = p_list.build_eclipse_data(family=1)
    assert "SWOF" in relperm_str
    assert "Simplified" in relperm_str  # Bad practice, testing for stuff in comments
    assert "a=1.5" in relperm_str


def test_comments_df():
    """Test that we support a tag column in the dataframe."""
    dframe = pd.DataFrame(
        columns=["SATNUM", "tag", "Nw", "Now", "ng", "nOG"],
        data=[[1, "thisisacomment", 2, 2, 2, 2]],
    )
    relperm_data = load_relperm_df(dframe)
    p_list = create_pyscal_list(relperm_data, h=0.2)
    relperm_str = p_list.build_eclipse_data(family=1)
    assert "thisisacomment" in relperm_str

    # tag vs comment in dataframe should not matter.
    dframe = pd.DataFrame(
        columns=["SATNUM", "comment", "Nw", "Now", "ng", "nOG"],
        data=[[1, "thisisacomment", 2, 2, 2, 2]],
    )
    relperm_data = load_relperm_df(dframe)
    p_list = create_pyscal_list(relperm_data, h=0.2)
    relperm_str = p_list.build_eclipse_data(family=1)
    assert relperm_str.count("thisisacomment") == 2

    # Check case insensitiveness:
    dframe = pd.DataFrame(
        columns=["SAtnUM", "coMMent", "Nw", "Now", "ng", "nOG"],
        data=[[1, "thisisacomment", 2, 2, 2, 2]],
    )
    relperm_data = load_relperm_df(dframe)
    p_list = create_pyscal_list(relperm_data, h=0.2)
    assert p_list.build_eclipse_data(family=1).count("thisisacomment") == 2

    # UTF-8 stuff:
    dframe = pd.DataFrame(
        columns=["SATNUM", "TAG", "Nw", "Now", "Ng", "Nog"],
        data=[[1, "æøå", 2, 2, 2, 2]],
    )
    relperm_data = load_relperm_df(dframe)
    p_list = create_pyscal_list(relperm_data, h=0.2)
    assert p_list.build_eclipse_data(family=1).count("æøå") == 2


def test_error_messages_pr_satnum():
    """When errors are somewhere in a big dataframe, we should
    provide hints to the user for where to look"""

    # A string instead of a number in SATNUM 2
    dframe = pd.DataFrame(
        columns=["SATNUM", "nw", "now"], data=[[1, 1, 1], [2, "foo", 1]]
    )
    with pytest.raises(ValueError, match="SATNUM 2"):
        create_pyscal_list(dframe, h=0.2)

    dframe = pd.DataFrame(
        columns=["SATNUM", "nw", "now"], data=[[1, 1, 1], [2, np.nan, 1]]
    )
    with pytest.raises(ValueError, match="SATNUM 2"):
        create_pyscal_list(dframe, h=0.2)

    # Mixed up order:
    dframe = pd.DataFrame(
        columns=["SATNUM", "nw", "now"], data=[[2, np.nan, 1], [1, 1, 1]]
    )
    with pytest.raises(ValueError, match="SATNUM 2"):
        create_pyscal_list(dframe, h=0.2)

    # Gasoil list
    dframe = pd.DataFrame(
        columns=["SATNUM", "ng", "nog"], data=[[1, 1, 1], [2, np.nan, 1]]
    )
    with pytest.raises(ValueError, match="SATNUM 2"):
        create_pyscal_list(dframe, h=0.2)

    # Gaswater list:
    dframe = pd.DataFrame(
        columns=["SATNUM", "ng", "nw"], data=[[1, 1, 1], [2, np.nan, 1]]
    )
    with pytest.raises(ValueError, match="SATNUM 2"):
        create_pyscal_list(dframe, h=0.2)

    # Wateroilgas list:
    dframe = pd.DataFrame(
        columns=["SATNUM", "nw", "now", "ng", "nog"],
        data=[[1, 1, 1, 1, 1], [2, np.nan, 1, 2, 3]],
    )
    with pytest.raises(ValueError, match="SATNUM 2"):
        create_pyscal_list(dframe, h=0.2)

    # SCAL rec list:
    dframe = pd.DataFrame(
        columns=["SATNUM", "CASE", "nw", "now", "ng", "nog"],
        data=[
            [1, "low", 1, 1, 1, 1],
            [1, "base", 1, 1, 1, 1],
            [1, "high", 1, 1, 1, 1],
            [2, "low", 1, 1, 2, 3],
            [2, "base", 2, 1, 2, 3],
            [2, "high", 3, 1, 2, 3],
        ],
    )
    # Perturb all entries in the dataframe:
    for rowidx in list(range(6)):
        for colidx in list(range(2, 6)):
            dframe_perturbed = dframe.copy()
            dframe_perturbed.iloc[rowidx, colidx] = np.nan
            satnum = dframe.iloc[rowidx, 0]
            case = dframe.iloc[rowidx, 1]
            # The error should hint both to SATNUM and to low/base/high
            with pytest.raises(ValueError, match=f"SATNUM {satnum}"):
                create_scal_recommendation_list(dframe_perturbed, h=0.2)
            with pytest.raises(ValueError, match=f"Problem with {case}"):
                create_scal_recommendation_list(dframe_perturbed, h=0.2)


def test_fast():
    """Test fast mode for SCALrecmmendation"""
    testdir = Path(__file__).absolute().parent

    scalrec_data = load_relperm_df(testdir / "data/scal-pc-input-example.xlsx")
    scalrec_list_fast = create_scal_recommendation_list(scalrec_data, fast=True)

    for item in scalrec_list_fast:
        assert item.fast
        assert item.low.fast
        assert item.base.fast
        assert item.high.fast

    wog_list_fast = scalrec_list_fast.interpolate(-0.5)
    for item in wog_list_fast:
        assert item.fast

    # WaterOilGas list
    dframe = pd.DataFrame(
        columns=["SATNUM", "nw", "now", "ng", "nog"],
        data=[
            [1, 2, 2, 2, 2],
            [2, 2, 2, 2, 2],
            [3, 2, 2, 2, 2],
        ],
    )
    relperm_data = load_relperm_df(dframe)
    p_list_fast = create_pyscal_list(relperm_data, h=0.2, fast=True)
    for item in p_list_fast:
        assert item.fast

    # GasOil list
    input_dframe = dframe[["SATNUM", "ng", "nog"]].copy()
    relperm_data = load_relperm_df(input_dframe)
    p_list_fast = create_pyscal_list(relperm_data, h=0.2, fast=True)
    for item in p_list_fast:
        assert item.fast

    # WaterOil list
    input_dframe = dframe[["SATNUM", "nw", "now"]].copy()
    relperm_data = load_relperm_df(input_dframe)
    p_list_fast = create_pyscal_list(relperm_data, h=0.2, fast=True)
    for item in p_list_fast:
        assert item.fast

    # GasWater list
    input_dframe = dframe[["SATNUM", "nw", "ng"]].copy()
    relperm_data = load_relperm_df(input_dframe)
    p_list_fast = create_pyscal_list(relperm_data, h=0.2, fast=True)
    for item in p_list_fast:
        assert item.fast

    # Testing with "fast" column in dataframe
    # Currently the choice is to only implement fast mode
    # as a global option. This column should do nothing now.
    # One could imagine it implemented
    # for individual SATNUM regions at a later stage
    dframe = pd.DataFrame(
        columns=["SATNUM", "nw", "now", "ng", "nog", "fast"],
        data=[
            [1, 2, 2, 2, 2, True],
            [2, 2, 2, 2, 2, False],
            [3, 2, 2, 2, 2, True],
        ],
    )
    relperm_data = load_relperm_df(dframe)
    assert "fast" in relperm_data
    p_list = create_pyscal_list(relperm_data, h=0.2)
    for item in p_list:
        assert not item.fast



"""SCALrecommendation, container for low, base and high WaterOilGas objects"""

import copy
from typing import Optional, Set, Type, Union

import numpy as np

from pyscal import GasWater, WaterOilGas, getLogger_pyscal
from pyscal.utils.interpolation import interpolate_go, interpolate_wo

logger = getLogger_pyscal(__name__)


class SCALrecommendation:
    """A SCAL recommendation consists of three OilWaterGas objects,
    tagged low, base and high.

    This container exists in order to to interpolation from -1 (low),
    through 0 (base) and to 1 (high).

    Args:
        low: An object representing the low case
        base: An object representing the base case
        high: An object representing the high case
        tag: A string that describes the recommendation. Optional.
    """

    def __init__(
        self,
        low: Union[WaterOilGas, GasWater],
        base: Union[WaterOilGas, GasWater],
        high: Union[WaterOilGas, GasWater],
        tag: Optional[str] = None,
        h: float = 0.01,
    ) -> None:
        """Set up a SCAL recommendation curve set from WaterOilGas objects

        Arguments:
            low: low case
            base: base case
            high: high case
            tag: Describes the recommendation. This string will be used
                as tag strings for the interpolants.
        """

        self.h: float = h
        self.tag: Optional[str] = tag
        self.low: Union[WaterOilGas, GasWater]
        self.base: Union[WaterOilGas, GasWater]
        self.high: Union[WaterOilGas, GasWater]
        self.type: Type

        if (
            isinstance(low, WaterOilGas)
            and isinstance(base, WaterOilGas)
            and isinstance(high, WaterOilGas)
        ):
            self.low = low
            self.base = base
            self.high = high
            self.type = WaterOilGas
        elif (
            isinstance(low, GasWater)
            and isinstance(base, GasWater)
            and isinstance(high, GasWater)
        ):
            self.low = low
            self.base = base
            self.high = high
            self.type = GasWater
        else:
            raise ValueError("Wrong arguments to SCALrecommendation")

        self.fast: bool = False
        if all([self.low.fast, self.base.fast, self.high.fast]):
            self.fast = True
        elif any([self.low.fast, self.base.fast, self.high.fast]):
            self.fast = self.low.fast = self.base.fast = self.high.fast = False
            logger.warning(
                (
                    "One or more of the low/base/high objects are set to be run in "
                    "fast mode, but not all. Fast mode set to false in all objects. "
                    "Code is run in normal mode."
                )
            )

    # User should add capillary pressure explicitly by calling add**
    # on the class objects, or run the following method to add the
    # same to all curves:
    def add_simple_J(
        self,
        a: float = 5.0,
        b: float = -1.5,
        poro_ref: float = 0.25,
        perm_ref: float = 100.0,
        drho: float = 300.0,
        g: float = 9.81,
    ) -> None:
        """Add (identical) simplified J-function to all water-oil
        curves in the SCAL recommendation set"""
        assert self.low.wateroil is not None
        assert self.base.wateroil is not None
        assert self.high.wateroil is not None
        self.low.wateroil.add_simple_J(
            a=a, b=b, poro_ref=poro_ref, perm_ref=perm_ref, drho=drho, g=g
        )
        self.base.wateroil.add_simple_J(
            a=a, b=b, poro_ref=poro_ref, perm_ref=perm_ref, drho=drho, g=g
        )
        self.high.wateroil.add_simple_J(
            a=a, b=b, poro_ref=poro_ref, perm_ref=perm_ref, drho=drho, g=g
        )

    def interpolate(
        self,
        parameter: float,
        parameter2: Optional[float] = None,
        h: Optional[float] = None,
    ) -> Union[WaterOilGas, GasWater]:
        """Interpolate between low, base and high

        Endpoints are located for input curves, and interpolated
        individually. Interpolation for the nonlinear part
        is done on a normalized interval between the endpoints

        Interpolation is linear in relperm-direction, and will thus not be
        linear in log-relperm-direction

        This method returns an WaterOilGas object which can be
        realized into printed tables. No attempt is made to
        parametrize the interpolant in L,E,T parameter space, or Corey-space.

        Args:
            parameter: Between -1 and 1, inclusive. -1 reproduces low/
                pessimistic curve, 0 gives base, 1 gives high/optimistic.
            parameter2: If not None, used for the gas-oil interpolation,
                enables having interpolation uncorrelated for WaterOil and
                GasOil. Ignored for GasWater (no warning).
            h: Saturation step length in generated tables. Does not
                need to be the same as the tables interpolation is done from.
        """

        gasparameter = parameter2 if parameter2 is not None else parameter

        # Either wateroil or gasoil can be None in the low, base, high
        # If they are None, it is a two-phase problem and we
        # should support this.
        do_gaswater = False
        do_wateroil = False
        do_gasoil = False
        if self.type == GasWater:
            do_gaswater = True
        elif self.type == WaterOilGas:
            do_wateroil = (
                self.base.wateroil is not None
                and self.low.wateroil is not None
                and self.high.wateroil is not None
            )

            do_gasoil = (
                self.base.gasoil is not None
                and self.low.gasoil is not None
                and self.high.gasoil is not None
            )

        if parameter2 is not None:
            if not do_gasoil:
                logger.warning("parameter2 is meaningless for water-oil only")
            if do_gaswater:
                logger.warning("parameter2 is meaningless for gas-water")

        # Initialize wateroil and gasoil curves to be filled with
        # interpolated curves:
        interpolant: Union[WaterOilGas, GasWater]
        tags: Set[str] = set()
        if do_wateroil or do_gaswater:
            assert self.low.wateroil is not None
            assert self.base.wateroil is not None
            assert self.high.wateroil is not None
            tags = tags.union(
                {
                    self.base.wateroil.tag,
                    self.low.wateroil.tag,
                    self.high.wateroil.tag,
                }
            )
        if do_gasoil:
            assert self.low.gasoil is not None
            assert self.base.gasoil is not None
            assert self.high.gasoil is not None
            tags = tags.union(
                {self.base.gasoil.tag, self.low.gasoil.tag, self.high.gasoil.tag}
            )
        tagstring = "\n".join(tags)
        if do_gaswater:
            interpolant = GasWater(h=h, tag=tagstring)
        else:
            interpolant = WaterOilGas(h=h, tag=tagstring)

        if do_wateroil or do_gaswater:
            tag = f"SCAL recommendation interpolation to {parameter}\n" + tagstring
            assert self.low.wateroil is not None
            assert self.base.wateroil is not None
            assert self.high.wateroil is not None
            if abs(parameter) > 1.0:
                raise ValueError(
                    f"Interpolation parameter must be in [-1,1], got {parameter}"
                )
            if np.isclose(parameter, 0.0):
                interpolant.wateroil = copy.deepcopy(self.base.wateroil)
                interpolant.wateroil.tag = tag
            elif np.isclose(parameter, -1.0):
                interpolant.wateroil = copy.deepcopy(self.low.wateroil)
                interpolant.wateroil.tag = tag
            elif np.isclose(parameter, 1.0):
                interpolant.wateroil = copy.deepcopy(self.high.wateroil)
                interpolant.wateroil.tag = tag
            elif parameter < 0.0:
                interpolant.wateroil = interpolate_wo(
                    self.base.wateroil,
                    self.low.wateroil,
                    -parameter,
                    h=h,
                    tag=tag,
                )
            elif parameter > 0.0:
                interpolant.wateroil = interpolate_wo(
                    self.base.wateroil,
                    self.high.wateroil,
                    parameter,
                    h=h,
                    tag=tag,
                )
        else:
            interpolant.wateroil = None

        if do_gasoil or do_gaswater:
            assert self.low.gasoil is not None
            assert self.base.gasoil is not None
            assert self.high.gasoil is not None
            tag = f"SCAL recommendation interpolation to {gasparameter}\n" + tagstring
            if abs(gasparameter) > 1.0:
                raise ValueError(
                    "Interpolation parameter for gas must "
                    f"be in [-1,1], got {gasparameter}"
                )
            if np.isclose(gasparameter, 0.0):
                interpolant.gasoil = copy.deepcopy(self.base.gasoil)
                interpolant.gasoil.tag = tag
            elif np.isclose(gasparameter, -1.0):
                interpolant.gasoil = copy.deepcopy(self.low.gasoil)
                interpolant.gasoil.tag = tag
            elif np.isclose(gasparameter, 1.0):
                interpolant.gasoil = copy.deepcopy(self.high.gasoil)
                interpolant.gasoil.tag = tag
            elif gasparameter < 0.0:
                interpolant.gasoil = interpolate_go(
                    self.base.gasoil,
                    self.low.gasoil,
                    -1 * gasparameter,
                    h=h,
                    tag=tag,
                )
            elif gasparameter > 0.0:
                interpolant.gasoil = interpolate_go(
                    self.base.gasoil,
                    self.high.gasoil,
                    gasparameter,
                    h=h,
                    tag=tag,
                )
        else:
            interpolant.gasoil = None

        interpolant.fast = self.fast

        return interpolant
